from typing import List, Optional, TypeVar

T = TypeVar("T")


def list_index(the_list: List[T], value: T) -> Optional[int]:
    if value not in the_list:
        return None
    return the_list.index(value)



from .disassembler import Disassembler
from .spimdisasm_disassembler import SpimdisasmDisassembler
from .null_disassembler import NullDisassembler

from ..util import options

__instance: Disassembler = NullDisassembler()
__initialized = False


def create_disassembler_instance(skip_version_check: bool, splat_version: str):
    global __instance
    global __initialized
    if options.opts.platform in ["n64", "psx", "ps2", "psp"]:
        __instance = SpimdisasmDisassembler()
        __initialized = True

        __instance.check_version(skip_version_check, splat_version)
        __instance.configure()
        return

    raise NotImplementedError("No disassembler for requested platform")


def get_instance() -> Disassembler:
    global __instance
    global __initialized
    if not __initialized:
        raise Exception("Disassembler instance not initialized")
        return None
    return __instance



from dataclasses import dataclass
import re
from typing import Dict, List, Optional, Set, TYPE_CHECKING

import spimdisasm

from intervaltree import IntervalTree
from ..disassembler import disassembler_instance
from pathlib import Path

# circular import
if TYPE_CHECKING:
    from ..segtypes.segment import Segment

from . import log, options, progress_bar

all_symbols: List["Symbol"] = []
all_symbols_dict: Dict[int, List["Symbol"]] = {}
all_symbols_ranges = IntervalTree()
ignored_addresses: Set[int] = set()
to_mark_as_defined: Set[str] = set()

# Initialize a spimdisasm context, used to store symbols and functions
spim_context = spimdisasm.common.Context()

TRUEY_VALS = ["true", "on", "yes", "y"]
FALSEY_VALS = ["false", "off", "no", "n"]

splat_sym_types = {"func", "jtbl", "jtbl_label", "label"}

ILLEGAL_FILENAME_CHARS = ["<", ">", ":", '"', "/", "\\", "|", "?", "*"]


def check_valid_type(typename: str) -> bool:
    if typename[0].isupper():
        return True

    if typename in splat_sym_types:
        return True

    if typename in disassembler_instance.get_instance().known_types():
        return True

    return False


def is_truey(str: str) -> bool:
    return str.lower() in TRUEY_VALS


def is_falsey(str: str) -> bool:
    return str.lower() in FALSEY_VALS


def add_symbol(sym: "Symbol"):
    all_symbols.append(sym)
    if sym.vram_start is not None:
        if sym.vram_start not in all_symbols_dict:
            all_symbols_dict[sym.vram_start] = []
        all_symbols_dict[sym.vram_start].append(sym)

    # For larger symbols, add their ranges to interval trees for faster lookup
    if sym.size > 4:
        all_symbols_ranges.addi(sym.vram_start, sym.vram_end, sym)


def to_cname(symbol_name: str) -> str:
    symbol_name = re.sub(r"[^0-9a-zA-Z_]", "_", symbol_name)

    if symbol_name[0] in ["0", "1", "2", "3", "4", "5", "6", "7", "8", "9"]:
        symbol_name = "_" + symbol_name

    return symbol_name


def handle_sym_addrs(
    path: Path, sym_addrs_lines: List[str], all_segments: "List[Segment]"
):
    def get_seg_for_name(name: str) -> Optional["Segment"]:
        for segment in all_segments:
            if segment.name == name:
                return segment
        return None

    def get_seg_for_rom(rom: int) -> Optional["Segment"]:
        for segment in all_segments:
            if segment.contains_rom(rom):
                return segment
        return None

    seen_symbols: Dict[str, "Symbol"] = dict()
    prog_bar = progress_bar.get_progress_bar(sym_addrs_lines)
    prog_bar.set_description(f"Loading symbols ({path.stem})")
    line: str
    for line_num, line in enumerate(prog_bar):
        line = line.strip()
        if not line == "" and not line.startswith("//"):
            comment_loc = line.find("//")
            line_main = line
            line_ext = ""

            if comment_loc != -1:
                line_ext = line[comment_loc + 2 :].strip()
                line_main = line[:comment_loc].strip()

            try:
                assert line.count(";") == 1, "Line must contain a single semi-colon"
                line_split = line_main.split("=")
                name = line_split[0].strip()
                addr = int(line_split[1].strip()[:-1], 0)
            except:
                log.parsing_error_preamble(path, line_num, line)
                log.write("Line must be of the form")
                log.write("<function_name> = <address>; // attr0:val0 attr1:val1 [...]")
                log.write("with <address> in hex preceded by 0x, or dec")
                log.write("")
                raise

            sym = Symbol(addr, given_name=name)

            ignore_sym = False
            if line_ext:
                for info in line_ext.split(" "):
                    if ":" in info:
                        if info.count(":") > 1:
                            log.parsing_error_preamble(path, line_num, line)
                            log.write(f"Too many ':'s in '{info}'")
                            log.error("")

                        attr_name, attr_val = info.split(":")
                        if attr_name == "":
                            log.parsing_error_preamble(path, line_num, line)
                            log.write(
                                f"Missing attribute name in '{info}', is there extra whitespace?"
                            )
                            log.error("")
                        if attr_val == "":
                            log.parsing_error_preamble(path, line_num, line)
                            log.write(
                                f"Missing attribute value in '{info}', is there extra whitespace?"
                            )
                            log.error("")

                        # Non-Boolean attributes
                        try:
                            if attr_name == "type":
                                if not check_valid_type(attr_val):
                                    log.parsing_error_preamble(path, line_num, line)
                                    log.write(
                                        f"Unrecognized symbol type in '{info}', it should be one of"
                                    )
                                    log.write(
                                        [
                                            *splat_sym_types,
                                            *spimdisasm.common.gKnownTypes,
                                        ]
                                    )
                                    log.write(
                                        "You may use a custom type that starts with a capital letter"
                                    )
                                    log.error("")
                                type = attr_val
                                sym.type = type
                                continue
                            if attr_name == "size":
                                size = int(attr_val, 0)
                                sym.given_size = size
                                continue
                            if attr_name == "rom":
                                rom_addr = int(attr_val, 0)
                                sym.rom = rom_addr
                                continue
                            if attr_name == "segment":
                                seg = get_seg_for_name(attr_val)
                                if seg is None:
                                    log.parsing_error_preamble(path, line_num, line)
                                    log.write(f"Cannot find segment '{attr_val}'")
                                    log.error("")
                                else:
                                    # Add segment to symbol
                                    sym.segment = seg
                                continue
                            if attr_name == "name_end":
                                sym.given_name_end = attr_val
                                continue
                            if attr_name == "filename":
                                sym.given_filename = attr_val
                                continue
                            if attr_name == "visibility":
                                sym.given_visibility = attr_val
                                continue
                            if attr_name == "function_owner":
                                sym.function_owner = attr_val
                                continue
                            if attr_name == "align":
                                align = int(attr_val, 0)

                                if align < 0:
                                    log.parsing_error_preamble(path, line_num, line)
                                    log.error(
                                        f"The given alignment for '{sym.name}' (0x{sym.vram_start:08X}) is negative."
                                    )
                                align_shift = (align & (-align)).bit_length() - 1
                                if (1 << align_shift) != align:
                                    log.parsing_error_preamble(path, line_num, line)
                                    log.error(
                                        f"The given alignment '0x{align:X}' for symbol '{sym.name}' (0x{sym.vram_start:08X}) is not a power of two."
                                    )
                                if sym.vram_start % align != 0:
                                    log.parsing_error_preamble(path, line_num, line)
                                    log.error(
                                        f"The symbol '{sym.name}' (0x{sym.vram_start:08X}) is not aligned already to the given alignment '0x{align:X}'."
                                    )

                                sym.given_align = align
                                continue
                        except:
                            log.parsing_error_preamble(path, line_num, line)
                            log.write(
                                f"value of attribute '{attr_name}' could not be read:"
                            )
                            log.write("")
                            raise

                        # Boolean attributes
                        tf_val = (
                            True
                            if is_truey(attr_val)
                            else False
                            if is_falsey(attr_val)
                            else None
                        )
                        if tf_val is None:
                            log.parsing_error_preamble(path, line_num, line)
                            log.write(
                                f"Invalid Boolean value '{attr_val}' for attribute '{attr_name}', should be one of"
                            )
                            log.write([*TRUEY_VALS, *FALSEY_VALS])
                            log.error("")
                        else:
                            if attr_name == "defined":
                                sym.defined = tf_val
                                continue
                            if attr_name == "extract":
                                sym.extract = tf_val
                                continue
                            if attr_name == "ignore":
                                ignore_sym = tf_val
                                continue
                            if attr_name == "force_migration":
                                sym.force_migration = tf_val
                                continue
                            if attr_name == "force_not_migration":
                                sym.force_not_migration = tf_val
                                continue
                            if attr_name == "allow_addend":
                                sym.allow_addend = tf_val
                                continue
                            if attr_name == "dont_allow_addend":
                                sym.dont_allow_addend = tf_val
                                continue
                            if attr_name == "can_reference":
                                sym.can_reference = tf_val
                                continue
                            if attr_name == "can_be_referenced":
                                sym.can_be_referenced = tf_val
                                continue
                            if attr_name == "allow_duplicated":
                                sym.allow_duplicated = True
                                continue
                            if attr_name == "use_non_matching_label":
                                sym.use_non_matching_label = tf_val

            if ignore_sym:
                if sym.given_size is None or sym.given_size == 0:
                    ignored_addresses.add(sym.vram_start)
                else:
                    spim_context.addBannedSymbolRangeBySize(
                        sym.vram_start, sym.given_size
                    )

                continue

            if sym.segment is None and sym.rom is not None:
                sym.segment = get_seg_for_rom(sym.rom)

            if sym.segment:
                sym.segment.add_symbol(sym)

            sym.user_declared = True

            if sym.name in seen_symbols:
                item = seen_symbols[sym.name]
                if not sym.allow_duplicated or not item.allow_duplicated:
                    log.parsing_error_preamble(path, line_num, line)
                    log.error(
                        f"Duplicate symbol detected! {sym.name} has already been defined at vram 0x{item.vram_start:08X}"
                    )

            if addr in all_symbols_dict:
                items = all_symbols_dict[addr]
                for item in items:
                    have_same_rom_addresses = sym.rom == item.rom
                    same_segment = sym.segment == item.segment

                    if have_same_rom_addresses and same_segment:
                        if not sym.allow_duplicated or not item.allow_duplicated:
                            log.parsing_error_preamble(path, line_num, line)
                            log.error(
                                f"Duplicate symbol detected! {sym.name} clashes with {item.name} defined at vram 0x{addr:08X}.\n  If this is intended, specify either a segment or a rom address for this symbol"
                            )

            if len(sym.filename) > 253 or any(
                c in ILLEGAL_FILENAME_CHARS for c in sym.filename
            ):
                log.parsing_error_preamble(path, line_num, line)
                log.error(
                    # sym.name is written on its own line so reading the error message is nicer because the sym name will be very long.
                    # Other lines have two spaces to make identation nicer and consistent
                    f"Ilegal symbol filename detected!\n"
                    f"  The symbol\n"
                    f"    {sym.name}\n"
                    f"  exceeds the 255 bytes filename limit that most OS imposes or uses illegal characters,\n"
                    f"  which will be a problem when writing the symbol to its own file.\n"
                    f"  To fix this specify a `filename` for this symbol, like `filename:func_{sym.vram_start:08X}`.\n"
                    f"  Make sure the filename does not exceed 253 bytes nor it contains any of the following characters:\n"
                    f"    {ILLEGAL_FILENAME_CHARS}"
                )

            seen_symbols[sym.name] = sym

            add_symbol(sym)


def initialize(all_segments: "List[Segment]"):
    global all_symbols
    global all_symbols_dict
    global all_symbols_ranges

    all_symbols = []
    all_symbols_dict = {}
    all_symbols_ranges = IntervalTree()

    # Manual list of func name / addrs
    for path in options.opts.symbol_addrs_paths:
        if path.exists():
            with open(path, encoding="utf-8") as f:
                sym_addrs_lines = f.readlines()
                handle_sym_addrs(path, sym_addrs_lines, all_segments)


def initialize_spim_context(all_segments: "List[Segment]") -> None:
    global_vrom_start = None
    global_vrom_end = None
    global_vram_start = options.opts.global_vram_start
    global_vram_end = options.opts.global_vram_end
    overlay_segments: Set[spimdisasm.common.SymbolsSegment] = set()

    spim_context.bannedSymbols |= ignored_addresses

    from ..segtypes.common.code import CommonSegCode

    global_segments_after_overlays: List[CommonSegCode] = []

    for segment in all_segments:
        if not isinstance(segment, CommonSegCode):
            # We only care about the VRAMs of code segments
            continue

        if segment.special_vram_segment:
            # Special segments which should not be accounted in the global VRAM calculation, like N64's IPL3
            continue

        if (
            not isinstance(segment.vram_start, int)
            or not isinstance(segment.vram_end, int)
            or not isinstance(segment.rom_start, int)
            or not isinstance(segment.rom_end, int)
        ):
            continue

        ram_id = segment.get_exclusive_ram_id()

        if ram_id is None:
            if global_vram_start is None:
                global_vram_start = segment.vram_start
            elif segment.vram_start < global_vram_start:
                global_vram_start = segment.vram_start

            if global_vram_end is None:
                global_vram_end = segment.vram_end
            elif global_vram_end < segment.vram_end:
                global_vram_end = segment.vram_end

                if len(overlay_segments) > 0:
                    # Global segment *after* overlay segments?
                    global_segments_after_overlays.append(segment)

            if global_vrom_start is None:
                global_vrom_start = segment.rom_start
            elif segment.rom_start < global_vrom_start:
                global_vrom_start = segment.rom_start

            if global_vrom_end is None:
                global_vrom_end = segment.rom_end
            elif global_vrom_end < segment.rom_end:
                global_vrom_end = segment.rom_end

        elif segment.vram_start != segment.vram_end:
            # Do not tell to spimdisasm about zero-sized segments.

            spim_segment = spim_context.addOverlaySegment(
                ram_id,
                segment.rom_start,
                segment.rom_end,
                segment.vram_start,
                segment.vram_end,
            )
            # Add the segment-specific symbols first
            for symbols_list in segment.seg_symbols.values():
                for sym in symbols_list:
                    add_symbol_to_spim_segment(spim_segment, sym)

            overlay_segments.add(spim_segment)

    if (
        global_vram_start is not None
        and global_vram_end is not None
        and global_vrom_start is not None
        and global_vrom_end is not None
    ):
        spim_context.changeGlobalSegmentRanges(
            global_vrom_start, global_vrom_end, global_vram_start, global_vram_end
        )

        overlaps_found = False
        # Check the vram range of the global segment does not overlap with any overlay segment
        for ovl_segment in overlay_segments:
            assert ovl_segment.vramStart <= ovl_segment.vramEnd, (
                f"{ovl_segment.vramStart:08X} {ovl_segment.vramEnd:08X}"
            )
            if (
                ovl_segment.vramEnd > global_vram_start
                and global_vram_end > ovl_segment.vramStart
            ):
                log.write(
                    f"Error: the vram range ([0x{ovl_segment.vramStart:08X}, 0x{ovl_segment.vramEnd:08X}]) of the non-global segment at rom address 0x{ovl_segment.vromStart:X} overlaps with the global vram range ([0x{global_vram_start:08X}, 0x{global_vram_end:08X}])",
                    status="warn",
                )
                overlaps_found = True
        if overlaps_found:
            log.write(
                "Many overlaps between non-global and global segments were found.",
            )
            log.write(
                "This is usually caused by missing `exclusive_ram_id` tags on segments that have a higher vram address than other `exclusive_ram_id`-tagged segments"
            )
            if len(global_segments_after_overlays) > 0:
                log.write(
                    "These segments are the main suspects for missing a `exclusive_ram_id` tag:",
                    status="warn",
                )
                for seg in global_segments_after_overlays:
                    log.write(f"    '{seg.name}', rom: 0x{seg.rom_start:06X}")
            else:
                log.write("No suspected segments??", status="warn")
            log.error("Stopping due to the above errors")

    # pass the global symbols to spimdisasm
    for segment in all_segments:
        if not isinstance(segment, CommonSegCode):
            # We only care about the VRAMs of code segments
            continue

        ram_id = segment.get_exclusive_ram_id()
        if ram_id is not None:
            continue

        for symbols_list in segment.seg_symbols.values():
            for sym in symbols_list:
                add_symbol_to_spim_segment(spim_context.globalSegment, sym)

    if global_vram_start and global_vram_end:
        # Pass global symbols to spimdisasm that are not part of any segment on the binary we are splitting (for psx and psp)
        for sym in all_symbols:
            if sym.segment is not None:
                # We already handled this symbol somewhere else
                continue

            if sym.vram_start < global_vram_start or sym.vram_end > global_vram_end:
                # Not global
                continue

            add_symbol_to_spim_segment(spim_context.globalSegment, sym)


def add_symbol_to_spim_segment(
    segment: spimdisasm.common.SymbolsSegment, sym: "Symbol"
) -> spimdisasm.common.ContextSymbol:
    if sym.type == "func":
        context_sym = segment.addFunction(
            sym.vram_start, isAutogenerated=not sym.user_declared, vromAddress=sym.rom
        )
    elif sym.type == "jtbl":
        context_sym = segment.addJumpTable(
            sym.vram_start, isAutogenerated=not sym.user_declared, vromAddress=sym.rom
        )
    elif sym.type == "jtbl_label":
        context_sym = segment.addJumpTableLabel(
            sym.vram_start, isAutogenerated=not sym.user_declared, vromAddress=sym.rom
        )
    elif sym.type == "label":
        context_sym = segment.addBranchLabel(
            sym.vram_start, isAutogenerated=not sym.user_declared, vromAddress=sym.rom
        )
    else:
        context_sym = segment.addSymbol(
            sym.vram_start, isAutogenerated=not sym.user_declared, vromAddress=sym.rom
        )
        if sym.type is not None:
            context_sym.type = sym.type

    if sym.user_declared:
        context_sym.isUserDeclared = True
    if sym.defined:
        context_sym.isDefined = True
    if sym.rom is not None:
        context_sym.vromAddress = sym.rom
    if sym.given_size is not None:
        context_sym.size = sym.size
    if sym.force_migration:
        context_sym.forceMigration = True
    if sym.force_not_migration:
        context_sym.forceNotMigration = True
    context_sym.functionOwnerForMigration = sym.function_owner
    if sym.allow_addend:
        context_sym.allowedToReferenceAddends = True
    if sym.dont_allow_addend:
        context_sym.notAllowedToReferenceAddends = True
    if sym.can_reference is not None:
        context_sym.allowedToReferenceSymbols = sym.can_reference
    if sym.can_be_referenced is not None:
        context_sym.allowedToBeReferenced = sym.can_be_referenced
    context_sym.setNameGetCallbackIfUnset(lambda _: sym.name)
    if sym.given_name_end:
        context_sym.nameEnd = sym.given_name_end
    if sym.given_visibility:
        context_sym.visibility = sym.given_visibility
    if sym.given_align:
        context_sym.setAlignment(sym.given_align)
    if sym.use_non_matching_label is not None:
        context_sym.useNonMatchingLabel = sym.use_non_matching_label

    return context_sym


def add_symbol_to_spim_section(
    section: spimdisasm.mips.sections.SectionBase, sym: "Symbol"
) -> spimdisasm.common.ContextSymbol:
    if sym.type == "func":
        context_sym = section.addFunction(
            sym.vram_start, isAutogenerated=not sym.user_declared, symbolVrom=sym.rom
        )
    elif sym.type == "jtbl":
        context_sym = section.addJumpTable(
            sym.vram_start, isAutogenerated=not sym.user_declared, symbolVrom=sym.rom
        )
    elif sym.type == "jtbl_label":
        context_sym = section.addJumpTableLabel(
            sym.vram_start, isAutogenerated=not sym.user_declared, symbolVrom=sym.rom
        )
    elif sym.type == "label":
        context_sym = section.addBranchLabel(
            sym.vram_start, isAutogenerated=not sym.user_declared, symbolVrom=sym.rom
        )
    else:
        context_sym = section.addSymbol(
            sym.vram_start, isAutogenerated=not sym.user_declared, symbolVrom=sym.rom
        )
        if sym.type is not None:
            context_sym.type = sym.type

    if sym.user_declared:
        context_sym.isUserDeclared = True
    if sym.defined:
        context_sym.isDefined = True
    if sym.rom is not None:
        context_sym.vromAddress = sym.rom
    if sym.given_size is not None:
        context_sym.size = sym.size
    if sym.force_migration:
        context_sym.forceMigration = True
    if sym.force_not_migration:
        context_sym.forceNotMigration = True
    context_sym.functionOwnerForMigration = sym.function_owner
    context_sym.setNameGetCallbackIfUnset(lambda _: sym.name)
    if sym.given_name_end:
        context_sym.nameEnd = sym.given_name_end
    if sym.given_visibility:
        context_sym.visibility = sym.given_visibility
    if sym.given_align:
        context_sym.setAlignment(sym.given_align)
    if sym.use_non_matching_label is not None:
        context_sym.useNonMatchingLabel = sym.use_non_matching_label

    return context_sym


# force_in_segment=True when the symbol belongs to this specific segment.
# force_in_segment=False when this symbol is just a reference.
def create_symbol_from_spim_symbol(
    segment: "Segment",
    context_sym: spimdisasm.common.ContextSymbol,
    *,
    force_in_segment: bool,
) -> "Symbol":
    in_segment = False

    sym_type = None
    if context_sym.type == spimdisasm.common.SymbolSpecialType.jumptable:
        in_segment = True
        sym_type = "jtbl"
    elif context_sym.type == spimdisasm.common.SymbolSpecialType.function:
        sym_type = "func"
    elif context_sym.type == spimdisasm.common.SymbolSpecialType.branchlabel:
        in_segment = True
        sym_type = "label"
    elif context_sym.type == spimdisasm.common.SymbolSpecialType.jumptablelabel:
        in_segment = True
        sym_type = "jtbl_label"

    if not in_segment:
        if (
            context_sym.overlayCategory is None
            and segment.get_exclusive_ram_id() is None
        ):
            in_segment = segment.contains_vram(context_sym.vram)
        elif context_sym.overlayCategory == segment.get_exclusive_ram_id():
            if context_sym.vromAddress is not None:
                in_segment = segment.contains_rom(context_sym.vromAddress)
            else:
                in_segment = segment.contains_vram(context_sym.vram)

    sym = segment.create_symbol(
        context_sym.vram, force_in_segment or in_segment, type=sym_type, reference=True
    )

    if sym.given_name is None and context_sym.name is not None:
        sym.given_name = context_sym.name

    # To keep the symbol name in sync between splat and spimdisasm
    context_sym.setNameGetCallback(lambda _: sym.name)

    if context_sym.size is not None:
        sym.given_size = context_sym.getSize()
    if context_sym.vromAddress is not None:
        sym.rom = context_sym.getVrom()
    if context_sym.isDefined:
        sym.defined = True
    if context_sym.referenceCounter > 0:
        sym.referenced = True

    return sym


def mark_c_funcs_as_defined():
    for symbol in all_symbols:
        if len(to_mark_as_defined) == 0:
            return
        sym_name = symbol.name
        if sym_name in to_mark_as_defined:
            symbol.defined = True
            to_mark_as_defined.remove(sym_name)


@dataclass
class Symbol:
    vram_start: int

    given_name: Optional[str] = None
    given_name_end: Optional[str] = None
    rom: Optional[int] = None
    type: Optional[str] = None
    given_size: Optional[int] = None
    segment: Optional["Segment"] = None

    defined: bool = False
    referenced: bool = False
    extract: bool = True
    user_declared: bool = False

    force_migration: bool = False
    force_not_migration: bool = False
    function_owner: Optional[str] = None

    allow_addend: bool = False
    dont_allow_addend: bool = False

    can_reference: Optional[bool] = None
    can_be_referenced: Optional[bool] = None

    linker_section: Optional[str] = None

    allow_duplicated: bool = False

    given_filename: Optional[str] = None
    given_visibility: Optional[str] = None

    given_align: Optional[int] = None

    use_non_matching_label: Optional[bool] = None

    _generated_default_name: Optional[str] = None
    _last_type: Optional[str] = None

    def __str__(self):
        return self.name

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Symbol):
            return False
        return self.vram_start == other.vram_start and self.segment == other.segment

    # https://stackoverflow.com/a/56915493/6292472
    def __hash__(self):
        return hash((self.vram_start, self.segment))

    def format_name(self, format: str) -> str:
        ret = format

        ret = ret.replace("$VRAM", f"{self.vram_start:08X}")

        if "$ROM" in ret:
            if not isinstance(self.rom, int):
                log.error(
                    f"Attempting to rom-name a symbol with no ROM address: {self.vram_start:08X} typed {self.type}"
                )
            ret = ret.replace("$ROM", f"{self.rom:X}")

        if "$SEG" in ret:
            if self.segment is None:
                # This probably is fine - we can't expect every symbol to have a segment. Fall back to just the ram address
                return f"{self.vram_start:X}"
            assert self.segment is not None
            ret = ret.replace("$SEG", self.segment.name)

        return ret

    @property
    def default_name(self) -> str:
        if self._generated_default_name is not None:
            if self.type == self._last_type:
                return self._generated_default_name

        if self.segment:
            if isinstance(self.rom, int):
                suffix = self.format_name(self.segment.symbol_name_format)
            else:
                suffix = self.format_name(self.segment.symbol_name_format_no_rom)
        else:
            if isinstance(self.rom, int):
                suffix = self.format_name(options.opts.symbol_name_format)
            else:
                suffix = self.format_name(options.opts.symbol_name_format_no_rom)

        if self.type == "func":
            prefix = "func"
        elif self.type == "jtbl":
            prefix = "jtbl"
        elif self.type in {"jtbl_label", "label"}:
            return f".L{suffix}"
        else:
            prefix = "D"

        self._last_type = self.type
        self._generated_default_name = f"{prefix}_{suffix}"
        return self._generated_default_name

    @property
    def rom_end(self):
        return None if not self.rom else self.rom + self.size

    @property
    def vram_end(self):
        return self.vram_start + self.size

    @property
    def name(self) -> str:
        return self.given_name if self.given_name else self.default_name

    @property
    def size(self) -> int:
        if self.given_size is not None:
            return self.given_size
        return 4

    @property
    def filename(self) -> str:
        if self.given_filename is not None:
            return self.given_filename
        return self.name

    def contains_vram(self, offset):
        return offset >= self.vram_start and offset < self.vram_end

    def contains_rom(self, offset):
        return offset >= self.rom and offset < self.rom_end


def get_all_symbols():
    global all_symbols
    return all_symbols


def reset_symbols():
    global all_symbols
    global all_symbols_dict
    global all_symbols_ranges
    global ignored_addresses
    global to_mark_as_defined
    all_symbols = []
    all_symbols_dict = {}
    all_symbols_ranges = IntervalTree()
    ignored_addresses = set()
    to_mark_as_defined = set()



from pathlib import Path
from typing import List

from ..linker_entry import LinkerEntry, LinkerWriter
from ..segment import Segment


class LinkerEntryOffset(LinkerEntry):
    def __init__(
        self,
        segment: Segment,
    ):
        super().__init__(segment, [], Path(), "linker_offset", "linker_offset", False)
        self.object_path = None

    def emit_entry(self, linker_writer: LinkerWriter):
        linker_writer._write_symbol(f"{self.segment.get_cname()}_OFFSET", ".")


class CommonSegLinker_offset(Segment):
    def get_linker_section_order(self) -> str:
        return ""

    def get_linker_section_linksection(self) -> str:
        return ""

    def get_linker_entries(self) -> List[LinkerEntry]:
        return [LinkerEntryOffset(self)]



from itertools import zip_longest
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union

from ...util import log, options
from ...util.color import unpack_color

from ..segment import Segment


VALID_SIZES = [0x20, 0x40, 0x80, 0x100, 0x200]


class N64SegPalette(Segment):
    require_unique_name = False

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if self.extract:
            if self.rom_end is None:
                log.error(
                    f"segment {self.name} needs to know where it ends; add a position marker [0xDEADBEEF] after it"
                )

            if isinstance(self.yaml, dict) and "size" in self.yaml:
                yaml_size: int = self.yaml["size"]
                if isinstance(self.rom_start, int) and isinstance(self.rom_end, int):
                    rom_len = self.rom_end - self.rom_start
                    if rom_len < yaml_size:
                        log.error(
                            f"Error: {self.name} has a `size` value of 0x{yaml_size:X}, but this is smaller than the actual rom size of the palette (0x{rom_len:X}, start: 0x{self.rom_start:X}, end: 0x{self.rom_end:X})"
                        )
                    elif rom_len > yaml_size:
                        log.error(
                            f"Warning: {self.name} has a `size` value of 0x{yaml_size:X}, but this is larger than the end of the palette (0x{rom_len:X}, start: 0x{self.rom_start:X}, end: 0x{self.rom_end:X})\n(hint add a 'bin' segment after this palette with address 0x{self.rom_end:X})",
                        )

                size = yaml_size
            else:
                assert self.rom_end is not None
                assert self.rom_start is not None
                actual_len = self.rom_end - self.rom_start

                hint_msg = "(hint: add a 'bin' segment after it or specify the size in the segment)"

                if actual_len > VALID_SIZES[-1]:
                    log.error(
                        f"Error: {self.name} (0x{actual_len:X} bytes) is too long, max 0x{VALID_SIZES[-1]:X})\n{hint_msg}"
                    )

                if actual_len not in VALID_SIZES:
                    log.error(
                        f"Error: {self.name} (0x{actual_len:X} bytes) is not a valid palette size ({', '.join(hex(s) for s in VALID_SIZES)})\n{hint_msg}"
                    )
                size = actual_len
        else:
            size = 0

        self.palette_size: int = size
        self.global_id: Optional[str] = (
            self.yaml.get("global_id") if isinstance(self.yaml, dict) else None
        )

    def get_cname(self) -> str:
        return super().get_cname() + "_pal"

    @staticmethod
    def parse_palette_bytes(data) -> List[Tuple[int, int, int, int]]:
        def iter_in_groups(iterable, n, fillvalue=None):
            args = [iter(iterable)] * n
            return zip_longest(*args, fillvalue=fillvalue)

        palette = []

        for a, b in iter_in_groups(data, 2):
            palette.append(unpack_color([a, b]))

        return palette

    def parse_palette(self, rom_bytes) -> List[Tuple[int, int, int, int]]:
        assert self.rom_start is not None
        data = rom_bytes[self.rom_start : self.rom_start + self.palette_size]

        return N64SegPalette.parse_palette_bytes(data)

    def should_split(self) -> bool:
        return self.extract and options.opts.is_mode_active("img")

    def out_path(self) -> Path:
        return options.opts.asset_path / self.dir / f"{self.name}.png"

    # TODO NEED NAMES...
    def get_linker_entries(self):
        from ..linker_entry import LinkerEntry

        return [
            LinkerEntry(
                self,
                [options.opts.asset_path / self.dir / f"{self.name}.png"],
                options.opts.asset_path / self.dir / f"{self.name}.pal",
                self.get_linker_section_order(),
                self.get_linker_section_linksection(),
                self.is_noload(),
            )
        ]

    @staticmethod
    def estimate_size(yaml: Union[Dict, List]) -> int:
        if isinstance(yaml, dict):
            if "size" in yaml:
                return int(yaml["size"])
        return 0x20



from .asm import CommonSegAsm


class CommonSegAsmtu(CommonSegAsm):
    def split(self, rom_bytes: bytes):
        if self.rom_start == self.rom_end:
            return

        if self.spim_section is None:
            return

        out_path = self.out_path()
        assert out_path is not None, str(self)

        self.split_as_asmtu_file(out_path)



#! /usr/bin/env python3

from __future__ import annotations

import argparse

import hashlib
import struct

import dataclasses

from pathlib import Path

import rabbitizer

# PSX EXE has the following layout
# header   ; 0x80 bytes
# padding  ; 0x780 bytes
# .rodata  ; variable length
# .text    ; variable length
# .data    ; variable length
# .sdata   ; variable length
# .bss     ; variable length, all zeroes
# .sbss    ; variable length, all zeroes

PAYLOAD_OFFSET = 0x800  # 0x80 byte header followed by 0x780 bytes of zeroes
WORD_SIZE_BYTES = 4

UNSUPPORTED_OPS = {
    # MIPS II
    "beql",
    "bgtzl",
    "blezl",
    "bnel",
    "ldc1",
    "ldc2",
    "ll",
    "sc",
    "sdc1",
    "sdc2",
    "sync",
    "teq",
    "tge",
    "tgei",
    "tgeiu",
    "tgeu",
    "tlt",
    "tltu",
    "tne",
    "tnei",
    # MIPS III
    "dadd",
    "daddi",
    "daddiu",
    "daddu",
    "dsub",
    "dsubu",
    "ld",
    "ldl",
    "ldr",
    "lld",
    "lwu",
    "scd",
    "sd",
    "sdl",
    "sdr",
    # MIPS IV
    "movn",
    "movz",
    "pref",
    "prefx",
}


def is_valid(insn) -> bool:
    if not insn.isValid():
        if insn.instrIdType.name in ("CPU_SPECIAL", "CPU_COP2"):
            return True
        else:
            return False

    opcode = insn.getOpcodeName()
    if opcode in UNSUPPORTED_OPS:
        return False

    return True


def try_find_text(
    rom_bytes, start_offset=PAYLOAD_OFFSET, valid_threshold=32
) -> tuple[int, int]:
    start = end = 0
    good_count = valid_count = 0

    in_text = False
    last_opcode = None

    words = struct.iter_unpack("<I", rom_bytes[start_offset:])
    for i, (word,) in enumerate(words):
        insn = rabbitizer.Instruction(word)

        if in_text:
            if not is_valid(insn):
                end = start_offset + i * WORD_SIZE_BYTES
                break
        else:
            if is_valid(insn):
                valid_count += 1

                opcode = insn.getOpcodeName()
                if last_opcode != opcode and opcode != "nop":
                    good_count += 1
            else:
                # reset
                good_count = valid_count = 0

            if good_count > valid_threshold:
                in_text = True
                start = start_offset + ((i + 1 - valid_count) * WORD_SIZE_BYTES)

            last_opcode = insn.getOpcodeName()

    return (start, end)


def try_get_gp(rom_bytes, start_offset, max_instructions=50) -> int:
    # $gp is set like this:
    # /* A7738 800B7138 0E801C3C */  lui        $gp, (0x800E0000 >> 16)
    # /* A773C 800B713C 90409C27 */  addiu      $gp, $gp, 0x4090
    gp = 0
    words = struct.iter_unpack("<I", rom_bytes[start_offset:])
    for i, (word,) in enumerate(words):
        if i > max_instructions:
            # give up
            break
        insn = rabbitizer.Instruction(word)
        if insn.getOpcodeName() == "lui" and insn.rt.name == "gp":
            gp = insn.getProcessedImmediate() << 16
        elif insn.getOpcodeName() == "addiu" and insn.rt.name == "gp":
            gp += insn.getProcessedImmediate()
            break
    return gp


def read_word(exe_bytes, offset) -> int:
    return struct.unpack("<I", exe_bytes[offset : offset + 4])[0]


@dataclasses.dataclass
class PsxExe:
    # Based on https://psx-spx.consoledev.net/cdromdrive/#filenameexe-general-purpose-executable
    entrypoint: int  # offset: 0x10
    initial_gp: int  # offset: 0x14
    destination_vram: int  # offset: 0x18
    payload_size: int  # offset: 0x1C
    # data_vram: int  # offset: 0x20
    # data_size: int  # offset: 0x24
    # bss_vram: int  # offset: 0x28
    # bss_size: int  # offset: 0x2C
    # initial_sp_base: int  # offset: 0x30
    # initial_sp_offset: int  # offset: 0x34

    text_start: int
    data_start: int

    size: int
    sha1: str

    @property
    def text_offset(self) -> int:
        return self.text_start

    @property
    def data_offset(self) -> int:
        return self.data_start

    @staticmethod
    def get_info(exe_path: Path, exe_bytes: bytes) -> PsxExe:
        entrypoint = read_word(exe_bytes, 0x10)
        destination_vram = read_word(exe_bytes, 0x18)
        payload_size = read_word(exe_bytes, 0x1C)

        text_start, data_start = try_find_text(exe_bytes)

        if text_start:
            entrypoint_rom = entrypoint + PAYLOAD_OFFSET - destination_vram
            initial_gp = try_get_gp(exe_bytes, entrypoint_rom)
        else:
            initial_gp = 0

        sha1 = hashlib.sha1(exe_bytes).hexdigest()

        return PsxExe(
            entrypoint,
            initial_gp,
            destination_vram,
            payload_size,
            text_start,
            data_start,
            len(exe_bytes),
            sha1,
        )


def main():
    parser = argparse.ArgumentParser(description="Gives information on PSX EXEs")
    parser.add_argument("exe", help="Path to an PSX EXE")

    args = parser.parse_args()

    exe_path = Path(args.exe)
    exe_bytes = exe_path.read_bytes()
    exe = PsxExe.get_info(exe_path, exe_bytes)

    print(f"Entrypoint: 0x{exe.entrypoint:08X}")

    print("Initial GP: ", end="")
    if exe.initial_gp != 0:
        print(f"0x{exe.initial_gp:08X}")
    else:
        print("No")

    print()
    print(f"Destination VRAM: 0x{exe.destination_vram:08X}")
    print(f"Payload size (without header): 0x{exe.payload_size:X}")

    print()
    print(f"Text binary offset (estimate): 0x{exe.text_offset:X}")
    if exe.data_offset != 0:
        print(f"Data binary offset (estimate): 0x{exe.data_offset:X}")

    print()
    print(f"File size: 0x{exe.size:X}")
    print(f"sha1: {exe.sha1}")


if __name__ == "__main__":
    main()



import os
import re
from pathlib import Path
from typing import Optional, Set, List

import rabbitizer
import spimdisasm

from ...util import log, options, symbols
from ...util.compiler import IDO
from ...util.symbols import Symbol

from .codesubsegment import CommonSegCodeSubsegment
from .rodata import CommonSegRodata


STRIP_C_COMMENTS_RE = re.compile(
    r'//.*?$|/\*.*?\*/|\'(?:\\.|[^\\\'])*\'|"(?:\\.|[^\\"])*"',
    re.DOTALL | re.MULTILINE,
)

C_FUNC_RE = re.compile(
    r"^(?:static\s+)?[^\s]+\s+([^\s(]+)\(([^;)]*)\)[^;]+?{", re.MULTILINE
)

C_GLOBAL_ASM_IDO_RE = re.compile(r"GLOBAL_ASM\(\"(\w+\/)*(\w+)\.s\"\)", re.MULTILINE)


class CommonSegC(CommonSegCodeSubsegment):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.defined_funcs: Set[str] = set()
        self.global_asm_funcs: Set[str] = set()
        self.global_asm_rodata_syms: Set[str] = set()

        self.file_extension = "c"

        self.use_gp_rel_macro = options.opts.use_gp_rel_macro_nonmatching

    @staticmethod
    def strip_c_comments(text):
        def replacer(match):
            s = match.group(0)
            if s.startswith("/"):
                return " "
            else:
                return s

        return re.sub(STRIP_C_COMMENTS_RE, replacer, text)

    @staticmethod
    def get_funcs_defined_in_c(c_file: Path) -> Set[str]:
        text = CommonSegC.strip_c_comments(c_file.read_text(encoding="utf-8"))

        return set(m.group(1) for m in C_FUNC_RE.finditer(text))

    @staticmethod
    def find_all_instances(string: str, sub: str):
        start = 0
        while True:
            start = string.find(sub, start)
            if start == -1:
                return
            yield start
            start += len(sub)

    @staticmethod
    def get_close_parenthesis(string: str, pos: int):
        paren_count = 0
        while True:
            cur_char = string[pos]
            if cur_char == "(":
                paren_count += 1
            elif cur_char == ")":
                if paren_count == 0:
                    return pos + 1
                else:
                    paren_count -= 1
            pos += 1

    @staticmethod
    def find_include_macro(text: str, macro_name: str):
        for pos in CommonSegC.find_all_instances(text, f"{macro_name}("):
            close_paren_pos = CommonSegC.get_close_parenthesis(
                text, pos + len(f"{macro_name}(")
            )
            macro_contents = text[pos:close_paren_pos]
            macro_args = macro_contents.split(",")
            if options.opts.use_legacy_include_asm:
                if len(macro_args) >= 3:
                    yield macro_args[2].strip(" )")
            else:
                if len(macro_args) >= 2:
                    yield macro_args[1].strip(" )")

    @staticmethod
    def find_include_asm(text: str):
        return CommonSegC.find_include_macro(text, "INCLUDE_ASM")

    @staticmethod
    def find_include_rodata(text: str):
        return CommonSegC.find_include_macro(text, "INCLUDE_RODATA")

    @staticmethod
    def get_global_asm_funcs(c_file: Path) -> Set[str]:
        text = CommonSegC.strip_c_comments(c_file.read_text(encoding="utf-8"))
        if options.opts.compiler == IDO:
            return set(m.group(2) for m in C_GLOBAL_ASM_IDO_RE.finditer(text))
        else:
            return set(CommonSegC.find_include_asm(text))

    @staticmethod
    def get_global_asm_rodata_syms(c_file: Path) -> Set[str]:
        text = CommonSegC.strip_c_comments(c_file.read_text(encoding="utf-8"))
        if options.opts.compiler == IDO:
            return set(m.group(2) for m in C_GLOBAL_ASM_IDO_RE.finditer(text))
        else:
            return set(CommonSegC.find_include_rodata(text))

    @staticmethod
    def is_text() -> bool:
        return True

    def get_section_flags(self) -> Optional[str]:
        return "ax"

    def out_path(self) -> Optional[Path]:
        return options.opts.src_path / self.dir / f"{self.name}.{self.file_extension}"

    def scan(self, rom_bytes: bytes):
        if (
            self.rom_start is not None
            and self.rom_end is not None
            and self.rom_start != self.rom_end
        ):
            path = self.out_path()
            if path:
                if options.opts.do_c_func_detection and path.exists():
                    # TODO run cpp?
                    self.defined_funcs = self.get_funcs_defined_in_c(path)
                    self.global_asm_funcs = self.get_global_asm_funcs(path)
                    self.global_asm_rodata_syms = self.get_global_asm_rodata_syms(path)
                    symbols.to_mark_as_defined.update(self.defined_funcs)
                    symbols.to_mark_as_defined.update(self.global_asm_funcs)
                    symbols.to_mark_as_defined.update(self.global_asm_rodata_syms)

            self.scan_code(rom_bytes)

    def split(self, rom_bytes: bytes):
        if self.is_auto_segment:
            if options.opts.make_full_disasm_for_code:
                self.split_as_asmtu_file(self.asm_out_path())
            return

        if self.rom_start != self.rom_end:
            asm_out_dir = options.opts.nonmatchings_path / self.dir
            matching_asm_out_dir = options.opts.matchings_path / self.dir

            self.print_file_boundaries()

            assert self.spim_section is not None and isinstance(
                self.spim_section.get_section(), spimdisasm.mips.sections.SectionText
            ), f"{self.name}, rom_start:{self.rom_start}, rom_end:{self.rom_end}"

            # We want to know if this C section has a corresponding rodata section so we can migrate its rodata
            rodata_section_type = ""
            rodata_spim_segment: Optional[spimdisasm.mips.sections.SectionRodata] = None
            if options.opts.migrate_rodata_to_functions:
                # We don't know if the rodata section is .rodata or .rdata, so we need to check both
                for sect in [".rodata", ".rdata"]:
                    rodata_sibling = self.siblings.get(sect)
                    if rodata_sibling is None:
                        continue

                    if rodata_sibling.is_generated:
                        continue

                    assert isinstance(rodata_sibling, CommonSegRodata), (
                        f"{rodata_sibling}, {rodata_sibling.type}"
                    )

                    if not rodata_sibling.type.startswith("."):
                        # Emit an error if we try to migrate the rodata symbols to functions if the rodata section is not prefixed with a dot
                        # (ie `- [0x1234, rodata, some_file]` instead of `- [0x1234, .rodata, some_file]`).
                        # Not prefixing the type with a dot would produce splat to both disassemble the rodata section to its own assembly file
                        # and to migrate the symbols to the corresponding functions, generating link-time errors and many headaches.
                        log.write(
                            f"\nProblem detected with the `{rodata_sibling.type}` section of the `{rodata_sibling.name}` file during rodata migration.",
                            status="warn",
                        )
                        log.write(
                            f"\t The `{rodata_sibling.type}` section was not prefixed with a dot, which is required for the rodata migration feature to work properly and avoid build errors due to duplicated symbols at link-time."
                        )
                        log.error(
                            f"\t To fix this, please prefix the section type with a dot (like `.{rodata_sibling.type}`)."
                        )

                    rodata_section_type = (
                        rodata_sibling.get_linker_section_linksection()
                    )

                    assert rodata_sibling.spim_section is not None, f"{rodata_sibling}"
                    assert isinstance(
                        rodata_sibling.spim_section.get_section(),
                        spimdisasm.mips.sections.SectionRodata,
                    )
                    rodata_spim_segment = rodata_sibling.spim_section.get_section()

                    # Stop searching
                    break

            # Precompute function-rodata pairings
            symbols_entries = (
                spimdisasm.mips.FunctionRodataEntry.getAllEntriesFromSections(
                    self.spim_section.get_section(), rodata_spim_segment
                )
            )

            is_new_c_file = False

            # Check and create the C file
            c_path = self.out_path()
            if c_path:
                if not c_path.exists() and options.opts.create_c_files:
                    self.create_c_file(asm_out_dir, c_path, symbols_entries)
                    is_new_c_file = True

                self.create_asm_dependencies_file(
                    c_path, asm_out_dir, is_new_c_file, symbols_entries
                )

            # Produce the asm files for functions
            for entry in symbols_entries:
                entry.sectionText = self.get_linker_section_linksection()
                entry.sectionRodata = rodata_section_type
                if entry.function is not None:
                    if (
                        entry.function.getNameUnquoted() in self.global_asm_funcs
                        or is_new_c_file
                        or options.opts.disassemble_all
                    ):
                        func_sym = self.get_symbol(
                            entry.function.vram,
                            in_segment=True,
                            type="func",
                            local_only=True,
                        )
                        assert func_sym is not None

                        if (
                            entry.function.getNameUnquoted()
                            not in self.global_asm_funcs
                            and options.opts.disassemble_all
                            and not is_new_c_file
                        ):
                            self.create_c_asm_file(
                                entry, matching_asm_out_dir, func_sym
                            )
                        else:
                            self.create_c_asm_file(entry, asm_out_dir, func_sym)

                else:
                    for spim_rodata_sym in entry.rodataSyms:
                        if (
                            spim_rodata_sym.getNameUnquoted()
                            in self.global_asm_rodata_syms
                            or is_new_c_file
                            or options.opts.disassemble_all
                        ):
                            rodata_sym = self.get_symbol(
                                spim_rodata_sym.vram, in_segment=True, local_only=True
                            )
                            assert rodata_sym is not None

                            self.create_unmigrated_rodata_file(
                                spim_rodata_sym, asm_out_dir, rodata_sym
                            )

            if options.opts.make_full_disasm_for_code:
                # Disable gpRelHack since this file is expected to be built with modern gas
                section = self.spim_section.get_section()
                old_value = section.getGpRelHack()
                section.setGpRelHack(False)

                if options.opts.platform == "ps2":
                    # Modern gas requires `$` on the special r5900 registers.
                    from rabbitizer import TrinaryValue

                    for func in section.symbolList:
                        assert isinstance(func, spimdisasm.mips.symbols.SymbolFunction)
                        for inst in func.instructions:
                            inst.flag_r5900UseDollar = TrinaryValue.TRUE

                self.split_as_asmtu_file(self.asm_out_path())

                if options.opts.platform == "ps2":
                    for func in section.symbolList:
                        assert isinstance(func, spimdisasm.mips.symbols.SymbolFunction)
                        for inst in func.instructions:
                            inst.flag_r5900UseDollar = TrinaryValue.FALSE

                section.setGpRelHack(old_value)

    def get_c_preamble(self):
        ret = []

        preamble = options.opts.generated_c_preamble
        ret.append(preamble)
        ret.append("")

        return ret

    def check_gaps_in_migrated_rodata(
        self,
        func: spimdisasm.mips.symbols.SymbolFunction,
        rodata_list: List[spimdisasm.mips.symbols.SymbolBase],
    ):
        for index in range(len(rodata_list) - 1):
            rodata_sym = rodata_list[index]
            next_rodata_sym = rodata_list[index + 1]

            if rodata_sym.vramEnd != next_rodata_sym.vram:
                log.write(
                    "\nA gap was detected in migrated rodata symbols!", status="warn"
                )
                log.write(
                    f"\t In function '{func.getNameUnquoted()}' (0x{func.vram:08X}), gap detected between '{rodata_sym.getNameUnquoted()}' (0x{rodata_sym.vram:08X}) and '{next_rodata_sym.getNameUnquoted()}' (0x{next_rodata_sym.vram:08X})"
                )
                log.write(
                    f"\t The address of the missing rodata symbol is 0x{rodata_sym.vramEnd:08X}"
                )
                log.write(
                    "\t Try to force the migration of that symbol with `force_migration:True` in the symbol_addrs.txt file; or avoid the migration of symbols around this address with `force_not_migration:True`"
                )

    def create_c_asm_file(
        self,
        func_rodata_entry: spimdisasm.mips.FunctionRodataEntry,
        out_dir: Path,
        func_sym: Symbol,
    ):
        outpath = out_dir / self.name / f"{func_sym.filename}.s"

        # Skip extraction if the file exists and the symbol is marked as extract=false
        if outpath.exists() and not func_sym.extract:
            return

        outpath.parent.mkdir(parents=True, exist_ok=True)

        with outpath.open("w", encoding="utf-8", newline="\n") as f:
            if options.opts.asm_inc_header:
                f.write(
                    options.opts.c_newline.join(options.opts.asm_inc_header.split("\n"))
                )

            named_registers_opt = rabbitizer.config.regNames_namedRegisters

            rabbitizer.config.regNames_namedRegisters = (
                options.opts.named_regs_for_c_funcs
            )
            func_rodata_entry.writeToFile(f)
            rabbitizer.config.regNames_namedRegisters = named_registers_opt

            if func_rodata_entry.function is not None:
                self.check_gaps_in_migrated_rodata(
                    func_rodata_entry.function, func_rodata_entry.rodataSyms
                )
                self.check_gaps_in_migrated_rodata(
                    func_rodata_entry.function, func_rodata_entry.lateRodataSyms
                )

        self.log(f"Disassembled {func_sym.filename} to {outpath}")

    def create_unmigrated_rodata_file(
        self,
        spim_rodata_sym: spimdisasm.mips.symbols.SymbolBase,
        out_dir: Path,
        rodata_sym: Symbol,
    ):
        outpath = out_dir / self.name / f"{rodata_sym.filename}.s"

        # Skip extraction if the file exists and the symbol is marked as extract=false
        if outpath.exists() and not rodata_sym.extract:
            return

        outpath.parent.mkdir(parents=True, exist_ok=True)

        with outpath.open("w", encoding="utf-8", newline="\n") as f:
            preamble = options.opts.generated_s_preamble
            if preamble:
                f.write(preamble + "\n")
            assert rodata_sym.linker_section is not None, rodata_sym.name
            f.write(f".section {rodata_sym.linker_section}\n\n")
            f.write(spim_rodata_sym.disassemble())

        self.log(f"Disassembled {rodata_sym.filename} to {outpath}")

    def get_c_line_include_macro(
        self,
        sym: Symbol,
        asm_out_dir: Path,
        macro_name: str,
    ) -> str:
        if options.opts.compiler == IDO:
            # IDO uses the asm processor to embeed assembly, and it doesn't require a special directive to include symbols
            asm_outpath = asm_out_dir / self.name / f"{sym.filename}.s"
            rel_asm_outpath = os.path.relpath(asm_outpath, options.opts.base_path)
            final_path = Path(rel_asm_outpath).as_posix()
            return f'#pragma GLOBAL_ASM("{final_path}")'

        if options.opts.use_legacy_include_asm:
            rel_asm_out_dir = asm_out_dir.relative_to(options.opts.nonmatchings_path)
            final_path = (rel_asm_out_dir / self.name).as_posix()
            return f'{macro_name}(const s32, "{final_path}", {sym.filename});'

        final_path = (asm_out_dir / self.name).as_posix()
        return f'{macro_name}("{final_path}", {sym.filename});'

    def get_c_lines_for_function(
        self,
        sym: Symbol,
        spim_sym: spimdisasm.mips.symbols.SymbolFunction,
        asm_out_dir: Path,
    ) -> List[str]:
        c_lines = []

        # Terrible hack to "auto-decompile" empty functions
        if (
            options.opts.auto_decompile_empty_functions
            and len(spim_sym.instructions) == 2
            and spim_sym.instructions[0].isReturn()
            and spim_sym.instructions[1].isNop()
        ):
            c_lines.append(f"void {spim_sym.getNameUnquoted()}(void) {{")
            c_lines.append("}")
        else:
            c_lines.append(
                self.get_c_line_include_macro(sym, asm_out_dir, "INCLUDE_ASM")
            )
        c_lines.append("")
        return c_lines

    def get_c_lines_for_rodata_sym(self, sym: Symbol, asm_out_dir: Path):
        c_lines = [self.get_c_line_include_macro(sym, asm_out_dir, "INCLUDE_RODATA")]
        c_lines.append("")
        return c_lines

    def create_c_file(
        self,
        asm_out_dir: Path,
        c_path: Path,
        symbols_entries: List[spimdisasm.mips.FunctionRodataEntry],
    ):
        c_lines = self.get_c_preamble()

        for entry in symbols_entries:
            if entry.function is not None:
                func_sym = self.get_symbol(
                    entry.function.vram,
                    in_segment=True,
                    type="func",
                    local_only=True,
                )
                assert func_sym is not None

                c_lines += self.get_c_lines_for_function(
                    func_sym, entry.function, asm_out_dir
                )
            else:
                for spim_rodata_sym in entry.rodataSyms:
                    rodata_sym = self.get_symbol(
                        spim_rodata_sym.vram, in_segment=True, local_only=True
                    )
                    assert rodata_sym is not None

                    c_lines += self.get_c_lines_for_rodata_sym(rodata_sym, asm_out_dir)

        c_path.parent.mkdir(parents=True, exist_ok=True)
        with c_path.open("w", encoding="utf-8", newline=options.opts.c_newline) as f:
            f.write("\n".join(c_lines))
        log.write(f"Wrote {self.name} to {c_path}")

    def create_asm_dependencies_file(
        self,
        c_path: Path,
        asm_out_dir: Path,
        is_new_c_file: bool,
        symbols_entries: List[spimdisasm.mips.FunctionRodataEntry],
    ):
        if not options.opts.create_asm_dependencies:
            return
        if (
            len(self.global_asm_funcs) + len(self.global_asm_rodata_syms)
        ) == 0 and not is_new_c_file:
            return

        assert self.spim_section is not None

        build_path = options.opts.build_path

        dep_path = build_path / c_path.with_suffix(".asmproc.d")
        dep_path.parent.mkdir(parents=True, exist_ok=True)
        with dep_path.open("w", newline=options.opts.c_newline) as f:
            if options.opts.use_o_as_suffix:
                o_path = build_path / c_path.with_suffix(".o")
            else:
                o_path = build_path / c_path.with_suffix(c_path.suffix + ".o")
            f.write(f"{o_path.as_posix()}:")
            depend_list = []
            for entry in symbols_entries:
                if entry.function is not None:
                    func_name = entry.function.getNameUnquoted()

                    if func_name in self.global_asm_funcs or is_new_c_file:
                        outpath = asm_out_dir / self.name / (func_name + ".s")
                        outpath.parent.mkdir(parents=True, exist_ok=True)

                        depend_list.append(outpath)
                        f.write(f" \\\n    {outpath.as_posix()}")
                else:
                    for rodata_sym in entry.rodataSyms:
                        rodata_name = rodata_sym.getNameUnquoted()

                        if rodata_name in self.global_asm_rodata_syms or is_new_c_file:
                            outpath = asm_out_dir / self.name / (rodata_name + ".s")
                            outpath.parent.mkdir(parents=True, exist_ok=True)

                            depend_list.append(outpath)
                            f.write(f" \\\n    {outpath.as_posix()}")

            f.write("\n")

            for depend_file in depend_list:
                f.write(f"{depend_file.as_posix()}:\n")



import os
import re
from functools import lru_cache
from pathlib import Path
from typing import Dict, List, OrderedDict, Set, Tuple, Union, Optional

from ..util import options, log

from .segment import Segment
from ..util.symbols import to_cname


# clean 'foo/../bar' to 'bar'
@lru_cache(maxsize=None)
def clean_up_path(path: Path) -> Path:
    path_resolved = path.resolve()
    base_resolved = options.opts.base_path.resolve()
    try:
        return path_resolved.relative_to(base_resolved)
    except ValueError:
        pass

    # If the path wasn't relative to the splat file, use the working directory instead
    cwd = Path(os.getcwd())
    try:
        return path_resolved.relative_to(cwd)
    except ValueError:
        pass

    # If it wasn't relative to that too, then just return the path as-is
    return path


def path_to_object_path(path: Path) -> Path:
    path = clean_up_path(path)
    if options.opts.use_o_as_suffix:
        full_suffix = ".o"
    else:
        full_suffix = path.suffix + ".o"

    if not str(path).startswith(str(options.opts.build_path)):
        path = options.opts.build_path / path
    return clean_up_path(path.with_suffix(full_suffix))


def write_file_if_different(path: Path, new_content: str):
    if path.exists():
        old_content = path.read_text(encoding="utf-8")
    else:
        old_content = ""

    if old_content != new_content:
        path.parent.mkdir(parents=True, exist_ok=True)
        with path.open("w", encoding="utf-8", newline=options.opts.c_newline) as f:
            f.write(new_content)


def get_segment_rom_start(cname: str) -> str:
    if options.opts.segment_symbols_style == "makerom":
        return f"_{cname}SegmentRomStart"
    return f"{cname}_ROM_START"


def get_segment_rom_end(cname: str) -> str:
    if options.opts.segment_symbols_style == "makerom":
        return f"_{cname}SegmentRomEnd"
    return f"{cname}_ROM_END"


def get_segment_vram_start(cname: str) -> str:
    if options.opts.segment_symbols_style == "makerom":
        return f"_{cname}SegmentStart"
    return f"{cname}_VRAM"


def get_segment_vram_end(cname: str) -> str:
    if options.opts.segment_symbols_style == "makerom":
        return f"_{cname}SegmentEnd"
    return f"{cname}_VRAM_END"


def convert_section_name_to_linker_format(section_type: str) -> str:
    assert section_type.startswith(".")
    if options.opts.segment_symbols_style == "makerom":
        if section_type == ".rodata":
            return "RoData"
        return section_type[1:].capitalize()

    return to_cname(section_type.upper())


def get_segment_section_start(segment_name: str, section_type: str) -> str:
    sec = convert_section_name_to_linker_format(section_type)
    if options.opts.segment_symbols_style == "makerom":
        return f"_{segment_name}Segment{sec}Start"
    return f"{segment_name}{sec}_START"


def get_segment_section_end(segment_name: str, section_type: str) -> str:
    sec = convert_section_name_to_linker_format(section_type)
    if options.opts.segment_symbols_style == "makerom":
        return f"_{segment_name}Segment{sec}End"
    return f"{segment_name}{sec}_END"


def get_segment_section_size(segment_name: str, section_type: str) -> str:
    sec = convert_section_name_to_linker_format(section_type)
    if options.opts.segment_symbols_style == "makerom":
        return f"_{segment_name}Segment{sec}Size"
    return f"{segment_name}{sec}_SIZE"


def get_segment_vram_end_symbol_name(segment: Segment) -> str:
    return get_segment_vram_end(segment.get_cname())


regex_data_segment_normalizer = re.compile(r"[^0-9a-zA-Z_]")


class LinkerEntry:
    def __init__(
        self,
        segment: Segment,
        src_paths: List[Path],
        object_path: Path,
        section_order: str,
        section_link: str,
        noload: bool = False,
    ):
        self.segment = segment
        self.src_paths = [clean_up_path(p) for p in src_paths]
        self.section_order = section_order
        self.section_link = section_link
        self.noload = noload
        self.bss_contains_common = segment.bss_contains_common
        self.object_path: Optional[Path] = path_to_object_path(object_path)

    @property
    def section_order_type(self) -> str:
        if self.section_order == ".rdata":
            return ".rodata"
        else:
            return self.section_order

    @property
    def section_link_type(self) -> str:
        if self.section_link == ".rdata":
            return ".rodata"
        else:
            return self.section_link

    def emit_symbol_for_data(self, linker_writer: "LinkerWriter"):
        if not options.opts.ld_generate_symbol_per_data_segment:
            return

        if self.object_path and self.section_link_type == ".data":
            path_cname = regex_data_segment_normalizer.sub(
                "_",
                str(self.segment.dir / self.segment.name)
                + ".".join(self.object_path.suffixes[:-1]),
            )
            linker_writer._write_symbol(path_cname, ".")

    def emit_path(self, linker_writer: "LinkerWriter"):
        assert self.object_path is not None, (
            f"{self.segment.name}, {self.segment.rom_start}"
        )

        if self.noload and self.bss_contains_common:
            linker_writer._write_object_path_section(
                self.object_path, f"{self.section_link} COMMON .scommon"
            )
        else:
            wildcard = "*" if options.opts.ld_wildcard_sections else ""

            linker_writer._write_object_path_section(
                self.object_path, f"{self.section_link}{wildcard}"
            )

    def emit_entry(self, linker_writer: "LinkerWriter"):
        self.emit_symbol_for_data(linker_writer)
        self.emit_path(linker_writer)


class LinkerWriter:
    def __init__(self, is_partial: bool = False):
        self.linker_discard_section: bool = options.opts.ld_discard_section
        self.sections_allowlist: List[str] = options.opts.ld_sections_allowlist
        self.sections_denylist: List[str] = options.opts.ld_sections_denylist
        # Used to store all the linker entries - build tools may want this information
        self.entries: List[LinkerEntry] = []
        self.dependencies_entries: List[LinkerEntry] = []

        self.buffer: List[str] = []
        self.header_symbols: Set[str] = set()

        self.is_partial: bool = is_partial

        self._indent_level = 0

        self._writeln("SECTIONS")
        self._begin_block()

        if not self.is_partial:
            self._writeln(f"__romPos = {options.opts.ld_rom_start};")

            if options.opts.ld_gp_expression is not None:
                self._writeln(f"_gp = {options.opts.ld_gp_expression};")
            elif options.opts.gp is not None:
                self._writeln(f"_gp = 0x{options.opts.gp:X};")

    # Write a series of statements which compute a symbol that represents the highest address among a list of segments' end addresses
    def write_max_vram_end_sym(self, symbol: str, overlays: List[Segment]):
        for segment in overlays:
            if segment == overlays[0]:
                self._writeln(
                    f"{symbol} = {get_segment_vram_end_symbol_name(segment)};"
                )
            else:
                self._writeln(
                    f"{symbol} = MAX({symbol}, {get_segment_vram_end_symbol_name(segment)});"
                )

    # Adds all the entries of a segment to the linker script buffer
    def add(self, segment: Segment, max_vram_syms: List[Tuple[str, List[Segment]]]):
        entries = segment.get_linker_entries()
        self.entries.extend(entries)
        self.dependencies_entries.extend(entries)

        seg_name = segment.get_cname()

        for sym, segs in max_vram_syms:
            self.write_max_vram_end_sym(sym, segs)

        if options.opts.ld_legacy_generation:
            self.add_legacy(segment, entries)
            return

        section_entries: OrderedDict[str, List[LinkerEntry]] = OrderedDict()
        for section_name in segment.section_order:
            if section_name in options.opts.section_order:
                section_entries[section_name] = []

        # Add all entries to section_entries
        prev_entry = None
        for entry in entries:
            if entry.section_order_type in section_entries:
                # Search for the very first section type
                # This is required in case the very first entry is a type that's not listed on section_order (like linker_offset) because it would be dropped
                prev_entry = entry.section_order_type
                break

        any_load = False
        any_noload = False
        for entry in entries:
            if entry.section_order_type in section_entries:
                if entry.section_order_type not in section_entries:
                    log.error(
                        f"\nError on linker script generation: section '{entry.section_order_type}' not found.\n  Make sure the section '{entry.section_order_type}' is listed on 'section_order' of the yaml options."
                    )
                section_entries[entry.section_order_type].append(entry)
            elif prev_entry is not None:
                # If this section is not present in section_order or section_order then pretend it is part of the last seen section, mainly for handling linker_offset
                if prev_entry not in section_entries:
                    log.error(
                        f"\nError on linker script generation: section '{prev_entry}' not found.\n  Make sure the section '{prev_entry}' is listed on 'section_order' of the yaml options."
                    )
                section_entries[prev_entry].append(entry)
            any_load = any_load or not entry.noload
            any_noload = any_noload or entry.noload
            prev_entry = entry.section_order_type

        if segment.ld_align_segment_start:
            self._write_symbol(
                "__romPos", f"ALIGN(__romPos, 0x{segment.ld_align_segment_start:X})"
            )
            self._write_symbol(".", f"ALIGN(., 0x{segment.ld_align_segment_start:X})")

        seg_rom_start = get_segment_rom_start(seg_name)
        self._write_symbol(seg_rom_start, "__romPos")

        is_first = True
        if any_load:
            # Only emit normal segment if there's at least one normal entry
            self._write_segment_sections(
                segment, seg_name, section_entries, noload=False, is_first=is_first
            )
            is_first = False

        if any_noload:
            # Only emit NOLOAD segment if there is at least one noload entry
            self._write_segment_sections(
                segment, seg_name, section_entries, noload=True, is_first=is_first
            )
            is_first = False

        self._end_segment(segment, all_bss=not any_load)

    def add_legacy(self, segment: Segment, entries: List[LinkerEntry]):
        seg_name = segment.get_cname()

        # To keep track which sections has been started
        started_sections: Dict[str, bool] = {
            section_name: False for section_name in options.opts.section_order
        }

        # Find where sections are last seen
        last_seen_sections: Dict[LinkerEntry, str] = {}
        for entry in reversed(entries):
            if (
                entry.section_order_type in options.opts.section_order
                and entry.section_order_type not in last_seen_sections.values()
            ):
                last_seen_sections[entry] = entry.section_order_type

        if segment.ld_align_segment_start:
            self._write_symbol(
                "__romPos", f"ALIGN(__romPos, 0x{segment.ld_align_segment_start:X})"
            )
            self._write_symbol(".", f"ALIGN(., 0x{segment.ld_align_segment_start:X})")

        seg_rom_start = get_segment_rom_start(seg_name)
        self._write_symbol(seg_rom_start, "__romPos")

        self._begin_segment(segment, seg_name, noload=False, is_first=True)

        i = 0
        for entry in entries:
            if entry.noload:
                break

            started = started_sections.get(entry.section_order_type, True)
            if not started:
                self._begin_section(seg_name, entry.section_order_type)
                started_sections[entry.section_order_type] = True

            entry.emit_entry(self)

            if entry in last_seen_sections:
                self._end_section(seg_name, entry.section_order_type, segment)

            i += 1

        if any(entry.noload for entry in entries):
            self._end_block()

            self._begin_segment(segment, seg_name, noload=True, is_first=False)

            for entry in entries[i:]:
                started = started_sections.get(entry.section_order_type, True)
                if not started:
                    self._begin_section(seg_name, entry.section_order_type)
                    started_sections[entry.section_order_type] = True

                entry.emit_entry(self)

                if entry in last_seen_sections:
                    self._end_section(seg_name, entry.section_order_type, segment)

        self._end_segment(segment, all_bss=False)

    def add_referenced_partial_segment(
        self, segment: Segment, max_vram_syms: List[Tuple[str, List[Segment]]]
    ):
        entries = segment.get_linker_entries()
        self.entries.extend(entries)

        segments_path = options.opts.ld_partial_build_segments_path
        assert segments_path is not None

        seg_name = segment.get_cname()

        for sym, segs in max_vram_syms:
            self.write_max_vram_end_sym(sym, segs)

        if segment.ld_align_segment_start:
            self._write_symbol(
                "__romPos", f"ALIGN(__romPos, 0x{segment.ld_align_segment_start:X})"
            )
            self._write_symbol(".", f"ALIGN(., 0x{segment.ld_align_segment_start:X})")

        seg_rom_start = get_segment_rom_start(seg_name)
        self._write_symbol(seg_rom_start, "__romPos")

        any_load = any(not e.noload for e in entries)
        is_first = True
        if any_load:
            # Only emit normal segment if there's at least one normal entry

            self._begin_segment(segment, seg_name, noload=False, is_first=is_first)

            for section_name in segment.section_order:
                if section_name not in options.opts.section_order:
                    continue
                if section_name == ".bss":
                    continue

                entry = LinkerEntry(
                    segment,
                    [],
                    segments_path / f"{seg_name}.o",
                    section_name,
                    section_name,
                    noload=False,
                )
                self.dependencies_entries.append(entry)
                entry.emit_entry(self)
            is_first = False

        if any(e.noload for e in entries):
            # Only emit NOLOAD segment if there is at least one noload entry

            if not is_first:
                self._end_block()

            self._begin_segment(segment, seg_name, noload=True, is_first=is_first)

            # Check if any section has the bss_contains_common option
            bss_contains_common = False
            for entry in entries:
                if entry.segment.bss_contains_common:
                    bss_contains_common = True
                    break

            entry = LinkerEntry(
                segment,
                [],
                segments_path / f"{seg_name}.o",
                ".bss",
                ".bss",
                noload=True,
            )
            entry.bss_contains_common = bss_contains_common
            self.dependencies_entries.append(entry)
            entry.emit_entry(self)

        self._end_segment(segment, all_bss=not any_load)

    def add_partial_segment(self, segment: Segment):
        entries = segment.get_linker_entries()
        self.entries.extend(entries)
        self.dependencies_entries.extend(entries)

        seg_name = segment.get_cname()

        section_entries: OrderedDict[str, List[LinkerEntry]] = OrderedDict()
        for section_name in segment.section_order:
            if section_name in options.opts.section_order:
                section_entries[section_name] = []

        # Add all entries to section_entries
        prev_entry = None
        for entry in entries:
            if entry.section_order_type in section_entries:
                if entry.section_order_type not in section_entries:
                    log.error(
                        f"\nError on linker script generation: section '{entry.section_order_type}' not found.\n  Make sure the section '{entry.section_order_type}' is listed on 'section_order' of the yaml options."
                    )
                section_entries[entry.section_order_type].append(entry)
            elif prev_entry is not None:
                # If this section is not present in section_order or section_order then pretend it is part of the last seen section, mainly for handling linker_offset
                if prev_entry not in section_entries:
                    log.error(
                        f"\nError on linker script generation: section '{prev_entry}' not found.\n  Make sure the section '{prev_entry}' is listed on 'section_order' of the yaml options."
                    )
                section_entries[prev_entry].append(entry)
            prev_entry = entry.section_order_type

        for section_name, entries in section_entries.items():
            if len(entries) == 0:
                continue
            first_entry = entries[0]

            self._begin_partial_segment(section_name, segment, first_entry.noload)

            self._begin_section(seg_name, section_name)

            for entry in entries:
                entry.emit_entry(self)

            self._end_section(seg_name, section_name, segment)

            self._end_partial_segment(section_name)

    def save_linker_script(self, output_path: Path):
        if len(self.sections_allowlist) > 0:
            address = " 0"
            if self.is_partial:
                address = ""
            for sect in self.sections_allowlist:
                self._writeln(f"{sect}{address} :")
                self._begin_block()
                self._writeln(f"*({sect});")
                self._end_block()

            self._writeln("")

        if self.linker_discard_section or len(self.sections_denylist) > 0:
            self._writeln("/DISCARD/ :")
            self._begin_block()
            for sect in self.sections_denylist:
                self._writeln(f"*({sect});")
            if self.linker_discard_section:
                self._writeln("*(*);")
            self._end_block()

        self._end_block()  # SECTIONS

        assert self._indent_level == 0

        write_file_if_different(output_path, "\n".join(self.buffer) + "\n")

    def save_symbol_header(self):
        path = options.opts.ld_symbol_header_path

        if path:
            write_file_if_different(
                path,
                "#ifndef _HEADER_SYMBOLS_H_\n"
                "#define _HEADER_SYMBOLS_H_\n"
                "\n"
                '#include "common.h"\n'
                "\n"
                + "".join(
                    f"extern Addr {symbol};\n" for symbol in sorted(self.header_symbols)
                )
                + "\n"
                "#endif\n",
            )

    def save_dependencies_file(self, output_path: Path, target_elf_path: Path):
        seen = dict.fromkeys(
            entry.object_path
            for entry in self.dependencies_entries
            if entry.object_path is not None
        )
        output = f"{clean_up_path(target_elf_path).as_posix()}:"
        for path in seen:
            output += f" \\\n    {path.as_posix()}"
        output += "\n"
        for path in seen:
            output += f"{path.as_posix()}:\n"
        if options.opts.ld_dependencies_include:
            deps = [
                path.with_suffix(".d").as_posix()
                for path in seen
                if path.suffix == ".o"
            ]
            if deps:
                output += f"-include {' '.join(deps)}\n"
        write_file_if_different(output_path, output)

    def _writeln(self, line: str):
        if len(line) == 0:
            self.buffer.append(line)
        else:
            self.buffer.append("    " * self._indent_level + line)

    def _begin_block(self):
        self._writeln("{")
        self._indent_level += 1

    def _end_block(self):
        self._indent_level -= 1
        self._writeln("}")

    def _write_symbol(self, symbol: str, value: Union[str, int]):
        symbol = to_cname(symbol)

        if isinstance(value, int):
            value = f"0x{value:X}"

        self._writeln(f"{symbol} = {value};")

        self.header_symbols.add(symbol)

    def _write_object_path_section(self, object_path: Path, section: str):
        self._writeln(f"{object_path.as_posix()}({section});")

    def _begin_segment(
        self, segment: Segment, seg_name: str, noload: bool, is_first: bool
    ):
        if (
            options.opts.ld_use_symbolic_vram_addresses
            and segment.vram_symbol is not None
        ):
            vram_str = segment.vram_symbol + " "
        else:
            vram_str = (
                f"0x{segment.vram_start:X} "
                if isinstance(segment.vram_start, int)
                else ""
            )

        addr_str = " "
        if is_first:
            addr_str += f"{vram_str}"
        if noload:
            seg_name += "_bss"
            addr_str += "(NOLOAD) "

        seg_vram_start = get_segment_vram_start(seg_name)
        self._write_symbol(seg_vram_start, f"ADDR(.{seg_name})")

        line = f".{seg_name}{addr_str}:"
        if not noload:
            seg_rom_start = get_segment_rom_start(seg_name)
            line += f" AT({seg_rom_start})"
        if options.opts.emit_subalign and segment.subalign is not None:
            line += f" SUBALIGN({segment.subalign})"

        self._writeln(line)
        self._begin_block()

        if segment.ld_fill_value is not None:
            self._writeln(f"FILL(0x{segment.ld_fill_value:08X});")

    def _end_segment(self, segment: Segment, all_bss=False):
        self._end_block()

        name = segment.get_cname()

        if not all_bss:
            self._writeln(f"__romPos += SIZEOF(.{name});")

        # Align directive
        if not options.opts.segment_end_before_align:
            if segment.align:
                self._writeln(f"__romPos = ALIGN(__romPos, {segment.align});")
                if options.opts.ld_align_segment_vram_end:
                    self._writeln(f". = ALIGN(., {segment.align});")

        seg_rom_end = get_segment_rom_end(name)
        self._write_symbol(seg_rom_end, "__romPos")
        self._write_symbol(get_segment_vram_end_symbol_name(segment), ".")

        # Align directive
        if options.opts.segment_end_before_align:
            if segment.align:
                self._writeln(f"__romPos = ALIGN(__romPos, {segment.align});")
                if options.opts.ld_align_segment_vram_end:
                    self._writeln(f". = ALIGN(., {segment.align});")

        self._writeln("")

    def _begin_partial_segment(self, section_name: str, segment: Segment, noload: bool):
        line = f"{section_name}"
        if noload:
            line += " (NOLOAD)"
        line += " :"
        if options.opts.emit_subalign and segment.subalign is not None:
            line += f" SUBALIGN({segment.subalign})"

        self._writeln(line)
        self._begin_block()

    def _end_partial_segment(self, section_name: str, all_bss=False):
        self._end_block()

        self._writeln("")

    def _begin_section(self, seg_name: str, cur_section: str) -> None:
        section_start = get_segment_section_start(seg_name, cur_section)
        self._write_symbol(section_start, ".")

    def _end_section(self, seg_name: str, cur_section: str, segment: Segment) -> None:
        section_start = get_segment_section_start(seg_name, cur_section)
        section_end = get_segment_section_end(seg_name, cur_section)
        section_size = get_segment_section_size(seg_name, cur_section)
        if options.opts.ld_align_section_vram_end and segment.align is not None:
            self._writeln(f". = ALIGN(., {segment.align});")
        self._write_symbol(section_end, ".")
        self._write_symbol(
            section_size,
            f"ABSOLUTE({section_end} - {section_start})",
        )

    def _write_segment_sections(
        self,
        segment: Segment,
        seg_name: str,
        section_entries: OrderedDict[str, List[LinkerEntry]],
        noload: bool,
        is_first: bool,
    ):
        if not is_first:
            self._end_block()

        self._begin_segment(segment, seg_name, noload=noload, is_first=is_first)

        for section_name, entries in section_entries.items():
            if len(entries) == 0:
                continue

            first_entry = entries[0]
            if first_entry.noload != noload:
                continue

            self._begin_section(seg_name, section_name)
            for entry in entries:
                entry.emit_entry(self)
            self._end_section(seg_name, section_name, segment)



from abc import ABC, abstractmethod
from typing import Optional

import spimdisasm

from ..util import options, symbols


class DisassemblerSection(ABC):
    @abstractmethod
    def disassemble(self):
        raise NotImplementedError("disassemble")

    @abstractmethod
    def analyze(self):
        raise NotImplementedError("analyze")

    @abstractmethod
    def set_comment_offset(self, rom_start: int):
        raise NotImplementedError("set_comment_offset")

    @abstractmethod
    def make_bss_section(
        self,
        rom_start,
        rom_end,
        vram_start,
        bss_end,
        name,
        segment_rom_start,
        exclusive_ram_id,
    ):
        raise NotImplementedError("make_bss_section")

    @abstractmethod
    def make_data_section(
        self,
        rom_start,
        rom_end,
        vram_start,
        name,
        rom_bytes,
        segment_rom_start,
        exclusive_ram_id,
    ):
        raise NotImplementedError("make_data_section")

    @abstractmethod
    def get_section(self):
        raise NotImplementedError("get_section")

    @abstractmethod
    def make_rodata_section(
        self,
        rom_start,
        rom_end,
        vram_start,
        name,
        rom_bytes,
        segment_rom_start,
        exclusive_ram_id,
    ):
        raise NotImplementedError("make_rodata_section")

    @abstractmethod
    def make_text_section(
        self,
        rom_start,
        rom_end,
        vram_start,
        name,
        rom_bytes,
        segment_rom_start,
        exclusive_ram_id,
    ):
        raise NotImplementedError("make_text_section")


class SpimdisasmDisassemberSection(DisassemblerSection):
    def __init__(self):
        self.spim_section: Optional[spimdisasm.mips.sections.SectionBase] = None

    def disassemble(self) -> str:
        assert self.spim_section is not None
        return self.spim_section.disassemble()

    def analyze(self):
        assert self.spim_section is not None
        self.spim_section.analyze()

    def set_comment_offset(self, rom_start: int):
        assert self.spim_section is not None
        self.spim_section.setCommentOffset(rom_start)

    def make_bss_section(
        self,
        rom_start: int,
        rom_end: int,
        vram_start: int,
        bss_end: int,
        name: str,
        segment_rom_start: int,
        exclusive_ram_id,
    ):
        self.spim_section = spimdisasm.mips.sections.SectionBss(
            symbols.spim_context,
            rom_start,
            rom_end,
            vram_start,
            bss_end,
            name,
            segment_rom_start,
            exclusive_ram_id,
        )

    def make_data_section(
        self,
        rom_start: int,
        rom_end: int,
        vram_start: int,
        name: str,
        rom_bytes: bytes,
        segment_rom_start: int,
        exclusive_ram_id,
    ):
        self.spim_section = spimdisasm.mips.sections.SectionData(
            symbols.spim_context,
            rom_start,
            rom_end,
            vram_start,
            name,
            rom_bytes,
            segment_rom_start,
            exclusive_ram_id,
        )

    def get_section(self) -> Optional[spimdisasm.mips.sections.SectionBase]:
        return self.spim_section

    def make_rodata_section(
        self,
        rom_start: int,
        rom_end: int,
        vram_start: int,
        name: str,
        rom_bytes: bytes,
        segment_rom_start: int,
        exclusive_ram_id,
    ):
        self.spim_section = spimdisasm.mips.sections.SectionRodata(
            symbols.spim_context,
            rom_start,
            rom_end,
            vram_start,
            name,
            rom_bytes,
            segment_rom_start,
            exclusive_ram_id,
        )

    def make_text_section(
        self,
        rom_start: int,
        rom_end: int,
        vram_start: int,
        name: str,
        rom_bytes: bytes,
        segment_rom_start: int,
        exclusive_ram_id,
    ):
        self.spim_section = spimdisasm.mips.sections.SectionText(
            symbols.spim_context,
            rom_start,
            rom_end,
            vram_start,
            name,
            rom_bytes,
            segment_rom_start,
            exclusive_ram_id,
        )

    def make_gcc_except_table_section(
        self,
        rom_start: int,
        rom_end: int,
        vram_start: int,
        name: str,
        rom_bytes: bytes,
        segment_rom_start: int,
        exclusive_ram_id,
    ):
        self.spim_section = spimdisasm.mips.sections.SectionGccExceptTable(
            symbols.spim_context,
            rom_start,
            rom_end,
            vram_start,
            name,
            rom_bytes,
            segment_rom_start,
            exclusive_ram_id,
        )


def make_disassembler_section() -> Optional[SpimdisasmDisassemberSection]:
    if options.opts.platform in ["n64", "psx", "ps2", "psp"]:
        return SpimdisasmDisassemberSection()

    raise NotImplementedError("No disassembler section for requested platform")
    return None


def make_text_section(
    rom_start: int,
    rom_end: int,
    vram_start: int,
    name: str,
    rom_bytes: bytes,
    segment_rom_start: int,
    exclusive_ram_id,
) -> DisassemblerSection:
    section = make_disassembler_section()
    assert section is not None
    section.make_text_section(
        rom_start,
        rom_end,
        vram_start,
        name,
        rom_bytes,
        segment_rom_start,
        exclusive_ram_id,
    )
    return section


def make_data_section(
    rom_start: int,
    rom_end: int,
    vram_start: int,
    name: str,
    rom_bytes: bytes,
    segment_rom_start: int,
    exclusive_ram_id,
) -> DisassemblerSection:
    section = make_disassembler_section()
    assert section is not None
    section.make_data_section(
        rom_start,
        rom_end,
        vram_start,
        name,
        rom_bytes,
        segment_rom_start,
        exclusive_ram_id,
    )
    return section


def make_rodata_section(
    rom_start: int,
    rom_end: int,
    vram_start: int,
    name: str,
    rom_bytes: bytes,
    segment_rom_start: int,
    exclusive_ram_id,
) -> DisassemblerSection:
    section = make_disassembler_section()
    assert section is not None
    section.make_rodata_section(
        rom_start,
        rom_end,
        vram_start,
        name,
        rom_bytes,
        segment_rom_start,
        exclusive_ram_id,
    )
    return section


def make_bss_section(
    rom_start: int,
    rom_end: int,
    vram_start: int,
    bss_end: int,
    name: str,
    segment_rom_start: int,
    exclusive_ram_id,
) -> DisassemblerSection:
    section = make_disassembler_section()
    assert section is not None
    section.make_bss_section(
        rom_start,
        rom_end,
        vram_start,
        bss_end,
        name,
        segment_rom_start,
        exclusive_ram_id,
    )
    return section


def make_gcc_except_table_section(
    rom_start: int,
    rom_end: int,
    vram_start: int,
    name: str,
    rom_bytes: bytes,
    segment_rom_start: int,
    exclusive_ram_id,
) -> DisassemblerSection:
    section = make_disassembler_section()
    assert section is not None
    section.make_gcc_except_table_section(
        rom_start,
        rom_end,
        vram_start,
        name,
        rom_bytes,
        segment_rom_start,
        exclusive_ram_id,
    )
    return section



from pathlib import Path

from ...util import options

from .segment import CommonSegment


class CommonSegHeader(CommonSegment):
    @staticmethod
    def is_data() -> bool:
        return True

    def should_split(self):
        return self.extract and options.opts.is_mode_active("code")

    @staticmethod
    def get_line(typ, data, comment):
        if typ == "ascii":
            text = data.decode("ASCII").strip()
            text = text.replace("\x00", "\\0")  # escape NUL chars
            dstr = '"' + text + '"'
        else:  # .word, .byte
            dstr = "0x" + data.hex().upper()

        dstr = dstr.ljust(20 - len(typ))

        return f".{typ} {dstr} /* {comment} */"

    def out_path(self) -> Path:
        return options.opts.asm_path / self.dir / f"{self.name}.s"

    def parse_header(self, rom_bytes):
        return []

    def split(self, rom_bytes):
        header_lines = self.parse_header(rom_bytes)

        src_path = self.out_path()
        src_path.parent.mkdir(parents=True, exist_ok=True)
        with src_path.open("w", encoding="utf-8", newline="\n") as f:
            f.write("\n".join(header_lines))
        self.log(f"Wrote {self.name} to {src_path}")

    @staticmethod
    def get_default_name(addr):
        return "header"



from . import linker_entry as linker_entry
from . import segment as segment

from . import common as common
from . import n64 as n64
from . import ps2 as ps2
from . import psx as psx
from . import psp as psp



from typing import List, Optional

from ...util import log

from .segment import CommonSegment
from ..segment import empty_statistics, Segment, SegmentStatistics


class CommonSegGroup(CommonSegment):
    def __init__(
        self,
        rom_start: Optional[int],
        rom_end: Optional[int],
        type: str,
        name: str,
        vram_start: Optional[int],
        args: list,
        yaml,
    ):
        super().__init__(
            rom_start,
            rom_end,
            type,
            name,
            vram_start,
            args=args,
            yaml=yaml,
        )

        self.subsegments: List[Segment] = self.parse_subsegments(yaml)

    def get_next_seg_start(self, i, subsegment_yamls) -> Optional[int]:
        j = i + 1
        while j < len(subsegment_yamls):
            ret, is_auto_segment = Segment.parse_segment_start(subsegment_yamls[j])
            if ret is not None:
                return ret
            j += 1

        # Fallback
        return self.rom_end

    def parse_subsegments(self, yaml) -> List[Segment]:
        ret: List[Segment] = []

        if not yaml or "subsegments" not in yaml:
            return ret

        prev_start: Optional[int] = -1
        last_rom_end = 0

        for i, subsegment_yaml in enumerate(yaml["subsegments"]):
            # endpos marker
            if isinstance(subsegment_yaml, list) and len(subsegment_yaml) == 1:
                continue

            typ = Segment.parse_segment_type(subsegment_yaml)
            start, is_auto_segment = Segment.parse_segment_start(subsegment_yaml)

            segment_class = Segment.get_class_for_type(typ)

            if start is None:
                # Attempt to infer the start address
                if i == 0:
                    # The start address of this segment is the start address of the group
                    start = self.rom_start
                else:
                    # The start address is the end address of the previous segment
                    start = last_rom_end

            # First, try to get the end address from the next segment's start address
            # Second, try to get the end address from the estimated size of this segment
            # Third, try to get the end address from the next segment with a start address
            end: Optional[int] = None
            if i < len(yaml["subsegments"]) - 1:
                end, end_is_auto_segment = Segment.parse_segment_start(
                    yaml["subsegments"][i + 1]
                )
            if start is not None and end is None:
                est_size = segment_class.estimate_size(subsegment_yaml)
                if est_size is not None:
                    end = start + est_size
            if end is None:
                end = self.get_next_seg_start(i, yaml["subsegments"])

            if start is not None and prev_start is not None and start < prev_start:
                log.error(
                    f"Error: Group segment {self.name} contains subsegments which are out of ascending rom order (0x{prev_start:X} followed by 0x{start:X})"
                )

            vram = None
            if start is not None:
                most_parent = self.get_most_parent()
                if (
                    most_parent.vram_start is not None
                    and most_parent.rom_start is not None
                ):
                    vram = most_parent.vram_start + start - most_parent.rom_start

            if segment_class.is_noload():
                # Pretend bss's rom address is after the last actual rom segment
                start = last_rom_end
                # and it has a rom size of zero
                end = last_rom_end

            segment: Segment = Segment.from_yaml(
                segment_class, subsegment_yaml, start, end, self, vram
            )
            if segment.special_vram_segment:
                self.special_vram_segment = True
            segment.is_auto_segment = is_auto_segment

            segment.index_within_group = len(ret)

            ret.append(segment)
            prev_start = start
            if end is not None:
                last_rom_end = end

        for i, seg in enumerate(ret):
            seg.index_within_group = i

        return ret

    @property
    def needs_symbols(self) -> bool:
        for seg in self.subsegments:
            if seg.needs_symbols:
                return True
        return False

    @property
    def statistics(self) -> SegmentStatistics:
        stats = empty_statistics()
        for sub in self.subsegments:
            for ty, info in sub.statistics.items():
                stats[ty] = stats[ty].merge(info)
        return stats

    def get_linker_entries(self):
        return [entry for sub in self.subsegments for entry in sub.get_linker_entries()]

    def scan(self, rom_bytes):
        for sub in self.subsegments:
            if sub.should_scan():
                sub.scan(rom_bytes)

    def split(self, rom_bytes):
        for sub in self.subsegments:
            if sub.should_split():
                sub.split(rom_bytes)

    def should_split(self) -> bool:
        return self.extract

    def should_scan(self) -> bool:
        return self.extract

    def cache(self):
        c = []

        for sub in self.subsegments:
            c.append(sub.cache())

        return c

    def get_subsegment_for_ram(self, addr: int) -> Optional[Segment]:
        for sub in self.subsegments:
            if sub.contains_vram(addr):
                return sub
        if isinstance(self.paired_segment, CommonSegGroup):
            for sub in self.paired_segment.subsegments:
                if sub.contains_vram(addr):
                    return sub
        return None

    def get_next_subsegment_for_ram(
        self, addr: int, current_subseg_index: Optional[int]
    ) -> Optional[Segment]:
        """
        Returns the first subsegment which comes after the specified address,
        or None in case this address belongs to the last subsegment of this group
        """

        start = current_subseg_index if current_subseg_index is not None else 0

        for sub in self.subsegments[start:]:
            if sub.vram_start is None:
                continue
            assert isinstance(sub.vram_start, int)
            if sub.vram_start > addr:
                return sub
        return None

    def pair_subsegments_to_other_segment(
        self,
        other_segment: "CommonSegGroup",
    ):
        # Pair cousins with the same name
        for segment in self.subsegments:
            for sibling in other_segment.subsegments:
                if segment.name == sibling.name:
                    # Make them reference each other
                    segment.siblings[sibling.get_linker_section_linksection()] = sibling
                    sibling.siblings[segment.get_linker_section_linksection()] = segment

                    if segment.is_text():
                        sibling.sibling = segment
                    elif sibling.is_text():
                        segment.sibling = sibling

                    break



from pathlib import Path
from typing import Optional

from ...util import log, options

from .segment import CommonSegment
from ..segment import SegmentType


class CommonSegBin(CommonSegment):
    @staticmethod
    def is_data() -> bool:
        return True

    def out_path(self) -> Optional[Path]:
        return options.opts.asset_path / self.dir / f"{self.name}.bin"

    def split(self, rom_bytes):
        path = self.out_path()
        assert path is not None
        path.parent.mkdir(parents=True, exist_ok=True)

        if self.rom_end is None:
            log.error(
                f"segment {self.name} needs to know where it ends; add a position marker [0xDEADBEEF] after it"
            )

        if self.size is None or self.size <= 0:
            log.error(f"Segment {self.name} has zero size.")

        assert isinstance(self.rom_start, int)
        assert isinstance(self.rom_end, int)
        path.write_bytes(rom_bytes[self.rom_start : self.rom_end])

        self.log(f"Wrote {self.name} to {path}")

    @property
    def statistics_type(self) -> SegmentType:
        stats_type = self.type
        if self.is_name_default():
            stats_type = "unk"
        return stats_type



from typing import Dict

from ..util import log

from ..segtypes.common.group import CommonSegGroup
from ..segtypes.n64.ci import N64SegCi
from ..segtypes.n64.palette import N64SegPalette

global_ids: Dict[str, N64SegPalette]


# Resolve Raster#palette and Palette#raster links
def initialize(all_segments):
    def collect_global_ids(segments):
        for segment in segments:
            if isinstance(segment, N64SegPalette):
                if segment.global_id is not None:
                    global_ids[segment.global_id] = segment

            if isinstance(segment, CommonSegGroup):
                collect_global_ids(segment.subsegments)

    def process(segments):
        raster_map: Dict[str, N64SegCi] = {}
        palette_map: Dict[str, N64SegPalette] = {}

        for segment in segments:
            if isinstance(segment, N64SegPalette):
                palette_map[segment.name] = segment

            if isinstance(segment, N64SegCi):
                raster_map[segment.name] = segment

            if isinstance(segment, CommonSegGroup):
                process(segment.subsegments)

        palettes_seen = set(palette_map.keys())

        for raster in raster_map.values():
            for pal_name in raster.palette_names:
                pal = global_ids.get(pal_name, None)
                if pal is not None:
                    global_ids_not_seen.discard(pal_name)
                    palettes_seen.discard(pal_name)
                    raster.palettes.append(pal)
                else:
                    pal = palette_map.get(pal_name, None)

                    if pal is not None:
                        palettes_seen.discard(pal_name)
                        raster.palettes.append(pal)
                    else:
                        log.error(
                            "Could not find palette: "
                            + pal_name
                            + ", referenced by raster: "
                            + raster.name
                        )

            # Resolve "." palette links
            for pal_name in palette_map:
                if pal_name.startswith(raster.name + "."):
                    pal = palette_map[pal_name]
                    raster.palettes.append(pal)
                    palettes_seen.discard(pal_name)

            if len(raster.palettes) == 0:
                log.error(f"Raster {raster.name} has no linked palettes")

        for pal_name in palettes_seen:
            pal = palette_map[pal_name]
            if pal.global_id is None:
                log.error(f"Palette {pal.name} has no linked rasters")

    global global_ids
    global_ids = {}

    collect_global_ids(all_segments)

    global_ids_not_seen = set(global_ids.keys())

    process(all_segments)

    if len(global_ids_not_seen) > 0:
        log.error(
            f"Found no ci links to palettes with global_ids: {', '.join(global_ids_not_seen)}"
        )



"""
N64 Vtx struct splitter
Dumps out Vtx as a .inc.c file.

Originally written by Mark Street (https://github.com/mkst)
"""

import struct
from pathlib import Path
from typing import Dict, List, Optional, Union

from ...util import options, log

from ..common.codesubsegment import CommonSegCodeSubsegment


class N64SegVtx(CommonSegCodeSubsegment):
    def __init__(
        self,
        rom_start: Optional[int],
        rom_end: Optional[int],
        type: str,
        name: str,
        vram_start: Optional[int],
        args: list,
        yaml,
    ):
        super().__init__(
            rom_start,
            rom_end,
            type,
            name,
            vram_start,
            args=args,
            yaml=yaml,
        )
        self.file_text: Optional[str] = None
        self.data_only = isinstance(yaml, dict) and yaml.get("data_only", False)

    def format_sym_name(self, sym) -> str:
        return sym.name

    def get_linker_section(self) -> str:
        return ".data"

    def out_path(self) -> Path:
        return options.opts.asset_path / self.dir / f"{self.name}.vtx.inc.c"

    def scan(self, rom_bytes: bytes):
        self.file_text = self.disassemble_data(rom_bytes)

    def disassemble_data(self, rom_bytes) -> str:
        assert isinstance(self.rom_start, int)
        assert isinstance(self.rom_end, int)
        assert isinstance(self.vram_start, int)

        vertex_data = rom_bytes[self.rom_start : self.rom_end]
        segment_length = len(vertex_data)
        if (segment_length) % 16 != 0:
            log.error(
                f"Error: Vtx segment {self.name} length ({segment_length}) is not a multiple of 16!"
            )

        lines = []
        if not self.data_only:
            lines.append(options.opts.generated_c_preamble)
            lines.append("")

        vertex_count = segment_length // 16
        sym = self.create_symbol(
            addr=self.vram_start, in_segment=True, type="data", define=True
        )

        if not self.data_only:
            lines.append(f"Vtx {self.format_sym_name(sym)}[{vertex_count}] = {{")

        for vtx in struct.iter_unpack(">hhhHhhBBBB", vertex_data):
            x, y, z, flg, t, c, r, g, b, a = vtx
            vtx_string = f"    {{{{{{ {x:5}, {y:5}, {z:5} }}, {flg}, {{ {t:5}, {c:5} }}, {{ {r:3}, {g:3}, {b:3}, {a:3} }}}}}},"
            if flg != 0:
                self.warn(f"Non-zero flag found in vertex data {self.name}!")
            lines.append(vtx_string)

        if not self.data_only:
            lines.append("};")

        # enforce newline at end of file
        lines.append("")
        return "\n".join(lines)

    def split(self, rom_bytes: bytes):
        out_path = self.out_path()
        if self.file_text and out_path:
            out_path.parent.mkdir(parents=True, exist_ok=True)

            with out_path.open("w", encoding="utf-8", newline="\n") as f:
                f.write(self.file_text)

    def should_scan(self) -> bool:
        return options.opts.is_mode_active("vtx")

    def should_split(self) -> bool:
        return self.extract and options.opts.is_mode_active("vtx")

    @staticmethod
    def estimate_size(yaml: Union[Dict, List]) -> Optional[int]:
        if isinstance(yaml, dict) and "length" in yaml:
            return yaml["length"] * 0x10
        return None



from ...util import log, options

from ..segment import Segment


class CommonSegDecompressor(Segment):
    def split(self, rom_bytes):
        out_dir = options.opts.asset_path / self.dir
        out_dir.mkdir(parents=True, exist_ok=True)

        if self.rom_end is None:
            log.error(
                f"segment {self.name} needs to know where it ends; add a position marker [0xDEADBEEF] after it"
            )

        out_path = out_dir / f"{self.name}.bin"
        assert isinstance(self.rom_start, int)
        assert isinstance(self.rom_end, int)

        self.log(f"Decompressing {self.name}")
        compressed_bytes = rom_bytes[self.rom_start : self.rom_end]
        decompressed_bytes = self.decompress(compressed_bytes)
        out_path.write_bytes(decompressed_bytes)
        self.log(f"Wrote {self.name} to {out_path}")

    def get_linker_entries(self):
        from ..linker_entry import LinkerEntry

        return [
            LinkerEntry(
                self,
                [options.opts.asset_path / self.dir / f"{self.name}.bin"],
                options.opts.asset_path
                / self.dir
                / f"{self.name}.{self.compression_type}",  # "MIO0" -> filename.MIO0.o
                self.get_linker_section_order(),
                self.get_linker_section_linksection(),
                self.is_noload(),
            )
        ]

    @property
    def compression_type(self) -> str:
        log.error(
            f"Segment {self.__class__.__name__} needs to define a compression type"
        )

    def decompress(self, compressed_bytes: bytes) -> bytes:
        log.error(
            f"Segment {self.__class__.__name__} needs to define a decompression method"
        )



from pathlib import Path

from .asm import CommonSegAsm

from ...util import options


class CommonSegHasm(CommonSegAsm):
    def asm_out_path(self) -> Path:
        if options.opts.hasm_in_src_path:
            return options.opts.src_path / self.dir / f"{self.name}.s"

        return super().asm_out_path()

    def scan(self, rom_bytes: bytes):
        if (
            self.rom_start is not None
            and self.rom_end is not None
            and self.rom_start != self.rom_end
        ):
            self.scan_code(rom_bytes, is_hasm=True)

    def split(self, rom_bytes: bytes):
        if self.rom_start == self.rom_end:
            return

        out_path = self.out_path()
        if out_path and not out_path.exists():
            self.split_as_asm_file(out_path)



from math import ceil

from . import options


# RRRRRGGG GGBBBBBA
def unpack_color(data):
    s = int.from_bytes(data[0:2], byteorder=options.opts.endianness)

    r = (s >> 11) & 0x1F
    g = (s >> 6) & 0x1F
    b = (s >> 1) & 0x1F
    a = (s & 1) * 0xFF

    r = ceil(0xFF * (r / 31))
    g = ceil(0xFF * (g / 31))
    b = ceil(0xFF * (b / 31))

    return r, g, b, a



__package_name__ = __name__

# Should be synced with pyproject.toml
__version__ = "0.40.0"
__author__ = "ethteck"

from . import util as util
from . import disassembler as disassembler
from . import platforms as platforms
from . import segtypes as segtypes

from . import scripts as scripts



import rabbitizer

from ..common.hasm import CommonSegHasm


class N64SegRsp(CommonSegHasm):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.instr_category = rabbitizer.InstrCategory.RSP



from pathlib import Path
from typing import Dict, List, Tuple, Type, Optional, Union

from n64img.image import Image
from ...util import log, options

from ..segment import Segment


class N64SegImg(Segment):
    @staticmethod
    def parse_dimensions(yaml: Union[Dict, List]) -> Tuple[int, int]:
        if isinstance(yaml, dict):
            return yaml["width"], yaml["height"]
        else:
            if len(yaml) < 5:
                log.error(f"Error: {yaml} is missing width and height parameters")
            return yaml[3], yaml[4]

    def __init__(
        self,
        rom_start: Optional[int],
        rom_end: Optional[int],
        type: str,
        name: str,
        vram_start: Optional[int],
        args: list,
        yaml,
        img_cls: Type[Image],
    ):
        super().__init__(
            rom_start,
            rom_end,
            type,
            name,
            vram_start,
            args=args,
            yaml=yaml,
        )

        if rom_start is None:
            log.error(f"Error: {type} segment {name} rom start could not be determined")

        self.n64img: Image = img_cls(b"", 0, 0)

        if isinstance(yaml, dict):
            self.n64img.flip_h = bool(yaml.get("flip_x", False))
            self.n64img.flip_v = bool(yaml.get("flip_y", False))

        self.width, self.height = self.parse_dimensions(yaml)

        self.n64img.width = self.width
        self.n64img.height = self.height

        self.image_type_in_extension = options.opts.image_type_in_extension

    def check_len(self) -> None:
        expected_len = int(self.n64img.size())
        assert isinstance(self.rom_start, int)
        assert isinstance(self.rom_end, int)

        actual_len = self.rom_end - self.rom_start
        if actual_len > expected_len:
            log.error(
                f"Error: {self.name} should end at 0x{self.rom_start + expected_len:X}, but it ends at 0x{self.rom_end:X}\n(hint: add a 'bin' segment after it)"
            )
        elif actual_len < expected_len:
            log.error(
                f"Error: {self.name} should end at 0x{self.rom_start + expected_len:X}, but it ends at 0x{self.rom_end:X}"
            )

    def out_path(self) -> Path:
        type_extension = f".{self.type}" if self.image_type_in_extension else ""

        return options.opts.asset_path / self.dir / f"{self.name}{type_extension}.png"

    def should_split(self) -> bool:
        return options.opts.is_mode_active("img")

    def split(self, rom_bytes):
        self.check_len()

        path = self.out_path()
        path.parent.mkdir(parents=True, exist_ok=True)

        if self.n64img.data == b"":
            self.n64img.data = rom_bytes[self.rom_start : self.rom_end]
        self.n64img.write(path)

        self.log(f"Wrote {self.name} to {path}")

    @staticmethod
    def estimate_size(yaml: Union[Dict, List]) -> int:
        width, height = N64SegImg.parse_dimensions(yaml)
        typ = Segment.parse_segment_type(yaml)

        if typ == "ci4" or typ == "i4" or typ == "ia4":
            return width * height // 2
        elif typ in ("ia16", "rgba16"):
            return width * height * 2
        elif typ == "rgba32":
            return width * height * 4
        else:
            return width * height



from pathlib import Path
from typing import Optional, List

import spimdisasm
import rabbitizer

from ...util import options, symbols, log

from .code import CommonSegCode

from ..segment import Segment, parse_segment_vram

from ...disassembler.disassembler_section import DisassemblerSection, make_text_section


# abstract class for c, asm, data, etc
class CommonSegCodeSubsegment(Segment):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        vram = parse_segment_vram(self.yaml)
        if vram is not None:
            self.vram_start = vram

        self.str_encoding: Optional[str] = (
            self.yaml.get("str_encoding", None) if isinstance(self.yaml, dict) else None
        )

        self.spim_section: Optional[DisassemblerSection] = None
        self.instr_category = rabbitizer.InstrCategory.CPU
        if options.opts.platform == "ps2":
            self.instr_category = rabbitizer.InstrCategory.R5900
        elif options.opts.platform == "psx":
            self.instr_category = rabbitizer.InstrCategory.R3000GTE
        elif options.opts.platform == "psp":
            self.instr_category = rabbitizer.InstrCategory.R4000ALLEGREX

        self.detect_redundant_function_end: Optional[bool] = (
            self.yaml.get("detect_redundant_function_end", None)
            if isinstance(self.yaml, dict)
            else None
        )

        self.is_hasm = False
        self.use_gp_rel_macro = options.opts.use_gp_rel_macro

    @property
    def needs_symbols(self) -> bool:
        return True

    def get_linker_section(self) -> str:
        return ".text"

    def configure_disassembler_section(
        self, disassembler_section: DisassemblerSection
    ) -> None:
        "Allows to configure the section before running the analysis on it"

        section = disassembler_section.get_section()

        section.isHandwritten = self.is_hasm
        section.instrCat = self.instr_category
        section.detectRedundantFunctionEnd = self.detect_redundant_function_end
        section.setGpRelHack(not self.use_gp_rel_macro)

    def scan_code(self, rom_bytes, is_hasm=False):
        self.is_hasm = is_hasm

        if self.is_auto_segment:
            return

        if not isinstance(self.rom_start, int):
            log.error(
                f"Segment '{self.name}' (type '{self.type}') requires a rom_start. Got '{self.rom_start}'"
            )

        # Supposedly logic error, not user error
        assert isinstance(self.rom_end, int), self.rom_end

        # Supposedly logic error, not user error
        segment_rom_start = self.get_most_parent().rom_start
        assert isinstance(segment_rom_start, int), segment_rom_start

        if not isinstance(self.vram_start, int):
            log.error(
                f"Segment '{self.name}' (type '{self.type}') requires a vram address. Got '{self.vram_start}'"
            )

        self.spim_section = make_text_section(
            self.rom_start,
            self.rom_end,
            self.vram_start,
            self.name,
            rom_bytes,
            segment_rom_start,
            self.get_exclusive_ram_id(),
        )

        assert self.spim_section is not None

        self.configure_disassembler_section(self.spim_section)

        self.spim_section.analyze()
        self.spim_section.set_comment_offset(self.rom_start)

        for func in self.spim_section.get_section().symbolList:
            assert isinstance(func, spimdisasm.mips.symbols.SymbolFunction)

            self.process_insns(func)

    def process_insns(
        self,
        func_spim: spimdisasm.mips.symbols.SymbolFunction,
    ):
        assert isinstance(self.parent, CommonSegCode)
        assert func_spim.vram is not None
        assert func_spim.vramEnd is not None
        assert self.spim_section is not None
        self.parent: CommonSegCode = self.parent

        symbols.create_symbol_from_spim_symbol(
            self.get_most_parent(), func_spim.contextSym, force_in_segment=False
        )

        # Gather symbols found by spimdisasm and create those symbols in splat's side
        for referenced_vram in func_spim.referencedVrams:
            context_sym = self.spim_section.get_section().getSymbol(
                referenced_vram, tryPlusOffset=False
            )
            if context_sym is not None:
                symbols.create_symbol_from_spim_symbol(
                    self.get_most_parent(), context_sym, force_in_segment=False
                )

        # Main loop
        for i, insn in enumerate(func_spim.instructions):
            if options.opts.platform == "ps2":
                from .c import CommonSegC
                from rabbitizer import TrinaryValue

                if isinstance(self, CommonSegC):
                    insn.flag_r5900UseDollar = TrinaryValue.FALSE
                else:
                    insn.flag_r5900UseDollar = TrinaryValue.TRUE
                insn.flag_r5900DisasmAsData = TrinaryValue.TRUE

            instr_offset = i * 4

            # update pointer accesses from this function
            if instr_offset in func_spim.instrAnalyzer.symbolInstrOffset:
                sym_address = func_spim.instrAnalyzer.symbolInstrOffset[instr_offset]

                context_sym = self.spim_section.get_section().getSymbol(sym_address)
                if context_sym is not None:
                    symbols.create_symbol_from_spim_symbol(
                        self.get_most_parent(), context_sym, force_in_segment=False
                    )

    def print_file_boundaries(self):
        if not self.show_file_boundaries or not self.spim_section:
            return

        assert isinstance(self.rom_start, int)

        for in_file_offset in self.spim_section.get_section().fileBoundaries:
            if not self.parent.reported_file_split:
                self.parent.reported_file_split = True

                # Look up for the last symbol in this boundary
                sym_addr = 0
                for sym in self.spim_section.get_section().symbolList:
                    symOffset = (
                        sym.inFileOffset - self.spim_section.get_section().inFileOffset
                    )
                    if in_file_offset == symOffset:
                        break
                    sym_addr = sym.vram

                print(
                    f"\nSegment {self.name}, symbol at vram {sym_addr:X} ends with extra nops, indicating a likely file split."
                )
                print(
                    "File split suggestions for this segment will follow in config yaml format:"
                )
            print(f"      - [0x{self.rom_start + in_file_offset:X}, {self.type}]")

    def should_scan(self) -> bool:
        return (
            options.opts.is_mode_active("code")
            and self.rom_start is not None
            and self.rom_end is not None
        )

    def should_split(self) -> bool:
        return (
            self.extract and options.opts.is_mode_active("code") and self.should_scan()
        )  # only split if the segment was scanned first

    def should_self_split(self) -> bool:
        return self.should_split()

    def get_asm_file_header(self) -> List[str]:
        ret = []

        ret.append('.include "macro.inc"')
        ret.append("")

        ret.extend(self.get_asm_file_extra_directives())

        preamble = options.opts.generated_s_preamble
        if preamble:
            ret.append(preamble)
            ret.append("")

        ret.append(self.get_section_asm_line())
        ret.append("")

        return ret

    def get_asm_file_extra_directives(self) -> List[str]:
        ret = []

        ret.append(".set noat")  # allow manual use of $at
        ret.append(".set noreorder")  # don't insert nops after branches
        if options.opts.add_set_gp_64:
            ret.append(".set gp=64")  # allow use of 64-bit general purpose registers
        ret.append("")

        return ret

    def asm_out_path(self) -> Path:
        return options.opts.asm_path / self.dir / f"{self.name}.s"

    def out_path(self) -> Optional[Path]:
        return self.asm_out_path()

    def split_as_asm_file(self, out_path: Optional[Path]):
        if self.spim_section is None:
            return

        if not out_path:
            return

        out_path.parent.mkdir(parents=True, exist_ok=True)

        self.print_file_boundaries()

        with out_path.open("w", encoding="utf-8", newline="\n") as f:
            # Write `.text` contents
            for line in self.get_asm_file_header():
                f.write(line + "\n")
            f.write(self.spim_section.disassemble())

    # Same as above but write all sections from siblings
    def split_as_asmtu_file(self, out_path: Path):
        out_path.parent.mkdir(parents=True, exist_ok=True)

        self.print_file_boundaries()

        with open(out_path, "w", encoding="utf-8", newline="\n") as f:
            for line in self.get_asm_file_header():
                f.write(line + "\n")

            # self.spim_section would be None if the current section was
            # declared `auto` in the yaml.
            # This can be useful when there are TUs with only data/rodata/bss/etc
            # sections but not text section.
            if self.spim_section is not None:
                # Write `.text` contents
                f.write(self.spim_section.disassemble())

            # Disassemble the siblings to this file by respecting the `section_order`
            for sect in self.section_order:
                if sect == self.get_linker_section_linksection():
                    continue

                sibling = self.siblings.get(sect)
                if sibling is None:
                    continue

                if (
                    isinstance(sibling, CommonSegCodeSubsegment)
                    and sibling.spim_section is not None
                    and (
                        not sibling.should_self_split()
                        or options.opts.make_full_disasm_for_code
                    )
                ):
                    f.write("\n")
                    f.write(f"{sibling.get_section_asm_line()}\n\n")
                    f.write(sibling.spim_section.disassemble())

            # Another loop to check anything that somehow may not be present in the `section_order`
            for sect, sibling in self.siblings.items():
                if sect == self.get_linker_section_linksection():
                    continue

                if sect in self.section_order:
                    # Already handled in the above loop
                    continue

                if (
                    isinstance(sibling, CommonSegCodeSubsegment)
                    and sibling.spim_section is not None
                    and (
                        not sibling.should_self_split()
                        or options.opts.make_full_disasm_for_code
                    )
                ):
                    f.write("\n")
                    f.write(f"{sibling.get_section_asm_line()}\n\n")
                    f.write(sibling.spim_section.disassemble())



from typing import Optional

from pathlib import Path

from ...util import options
from ..common.lib import CommonSegment
from ..segment import parse_segment_vram


class CommonSegO(CommonSegment):
    def __init__(
        self,
        rom_start: Optional[int],
        rom_end: Optional[int],
        type: str,
        name: str,
        vram_start: Optional[int],
        args: list,
        yaml,
    ):
        super().__init__(
            rom_start,
            rom_end,
            type,
            name,
            vram_start,
            args=args,
            yaml=yaml,
        )

        vram = parse_segment_vram(self.yaml)
        if vram is not None:
            self.vram_start = vram

        if isinstance(yaml, dict):
            self.section = yaml.get("section", ".text")
        else:
            if len(args) > 0:
                self.section = args[0]
            else:
                self.section = ".text"

        self.extract = False

    def get_linker_section(self) -> str:
        return self.section

    def out_path(self) -> Optional[Path]:
        out_path = options.opts.o_path / f"{self.name}"
        return out_path



from dataclasses import dataclass
from typing import Dict

import spimdisasm

from . import log, options, symbols, progress_bar


@dataclass
class Reloc:
    rom_address: int
    reloc_type: str
    symbol_name: str

    addend: int = 0


all_relocs: Dict[int, Reloc] = {}


def add_reloc(reloc: Reloc):
    all_relocs[reloc.rom_address] = reloc


def initialize():
    global all_relocs

    all_relocs = {}

    for path in options.opts.reloc_addrs_paths:
        if not path.exists():
            continue

        with path.open(encoding="utf-8") as f:
            sym_addrs_lines = f.readlines()

        prog_bar = progress_bar.get_progress_bar(sym_addrs_lines)
        prog_bar.set_description(f"Loading relocs ({path.stem})")
        line: str
        for line_num, line in enumerate(prog_bar):
            # Allow comments
            line = line.split("//")[0].strip()

            if line == "":
                continue

            rom_addr = None
            reloc_type = None
            symbol_name = None
            addend = None

            for info in line.split(" "):
                if ":" not in info:
                    continue

                if info.count(":") > 1:
                    log.parsing_error_preamble(path, line_num, line)
                    log.write(f"Too many ':'s in '{info}'")
                    log.error("")

                attr_name, attr_val = info.split(":")
                if attr_name == "":
                    log.parsing_error_preamble(path, line_num, line)
                    log.write(
                        f"Missing attribute name in '{info}', is there extra whitespace?"
                    )
                    log.error("")
                if attr_val == "":
                    log.parsing_error_preamble(path, line_num, line)
                    log.write(
                        f"Missing attribute value in '{info}', is there extra whitespace?"
                    )
                    log.error("")

                # Non-Boolean attributes
                try:
                    if attr_name == "rom":
                        rom_addr = int(attr_val, 0)
                        continue
                    if attr_name == "reloc":
                        reloc_type = attr_val
                        continue
                    if attr_name == "symbol":
                        symbol_name = attr_val
                        continue
                    if attr_name == "addend":
                        addend = int(attr_val, 0)
                        continue
                except:
                    log.parsing_error_preamble(path, line_num, line)
                    log.write(f"value of attribute '{attr_name}' could not be read:")
                    log.write("")
                    raise

            if rom_addr is None:
                log.parsing_error_preamble(path, line_num, line)
                log.error("Missing required 'rom' attribute for reloc")
            if reloc_type is None:
                log.parsing_error_preamble(path, line_num, line)
                log.error("Missing required 'reloc' attribute for reloc")
            if symbol_name is None:
                log.parsing_error_preamble(path, line_num, line)
                log.error("Missing required 'symbol' attribute for reloc")

            reloc = Reloc(rom_addr, reloc_type, symbol_name)
            if addend is not None:
                reloc.addend = addend

            if reloc.rom_address in all_relocs:
                log.parsing_error_preamble(path, line_num, line)
                log.error(
                    f"Duplicated 'rom' address for reloc: 0x{reloc.rom_address:X}"
                )
            add_reloc(reloc)


def initialize_spim_context():
    for rom_address, reloc in all_relocs.items():
        reloc_type = spimdisasm.common.RelocType.fromStr(reloc.reloc_type)

        if reloc_type is None:
            log.error(
                f"Reloc type '{reloc.reloc_type}' is not valid. Rom address: 0x{rom_address:X}"
            )

        symbols.spim_context.addGlobalReloc(
            rom_address, reloc_type, reloc.symbol_name, addend=reloc.addend
        )



"""
N64 f3dex display list splitter
Dumps out Gfx[] as a .inc.c file.
"""

import re
from typing import Dict, List, Optional, Union

from pathlib import Path

from pygfxd import (
    gfxd_buffer_to_string,
    gfxd_cimg_callback,
    gfxd_dl_callback,
    gfxd_endian,
    gfxd_execute,
    gfxd_input_buffer,
    gfxd_light_callback,
    gfxd_lookat_callback,
    gfxd_macro_dflt,
    gfxd_macro_fn,
    gfxd_mtx_callback,
    gfxd_output_buffer,
    gfxd_printf,
    gfxd_puts,
    gfxd_target,
    gfxd_timg_callback,
    gfxd_tlut_callback,
    gfxd_vp_callback,
    gfxd_vtx_callback,
    gfxd_zimg_callback,
    GfxdEndian,
    gfxd_f3d,
    gfxd_f3db,
    gfxd_f3dex,
    gfxd_f3dexb,
    gfxd_f3dex2,
)

from ...util import log, options
from ...util.log import error

from ..common.codesubsegment import CommonSegCodeSubsegment

from ...util import symbols

LIGHTS_RE = re.compile(r"\*\(Lightsn \*\)0x[0-9A-F]{8}")


class N64SegGfx(CommonSegCodeSubsegment):
    def __init__(
        self,
        rom_start: Optional[int],
        rom_end: Optional[int],
        type: str,
        name: str,
        vram_start: Optional[int],
        args: list,
        yaml,
    ):
        super().__init__(
            rom_start,
            rom_end,
            type,
            name,
            vram_start,
            args=args,
            yaml=yaml,
        )
        self.file_text = None
        self.data_only = isinstance(yaml, dict) and yaml.get("data_only", False)
        self.in_segment = not isinstance(yaml, dict) or yaml.get("in_segment", True)

    def format_sym_name(self, sym) -> str:
        return sym.name

    def get_linker_section(self) -> str:
        return ".data"

    def out_path(self) -> Path:
        return options.opts.asset_path / self.dir / f"{self.name}.gfx.inc.c"

    def scan(self, rom_bytes: bytes):
        self.file_text = self.disassemble_data(rom_bytes)

    def get_gfxd_target(self):
        opt = options.opts.gfx_ucode

        if opt == "f3d":
            return gfxd_f3d
        elif opt == "f3db":
            return gfxd_f3db
        elif opt == "f3dex":
            return gfxd_f3dex
        elif opt == "f3dexb":
            return gfxd_f3dexb
        elif opt == "f3dex2":
            return gfxd_f3dex2
        else:
            log.error(f"Unknown target {opt}")

    def tlut_handler(self, addr, idx, count):
        sym = self.create_symbol(
            addr=addr, in_segment=self.in_segment, type="data", reference=True
        )
        gfxd_printf(self.format_sym_name(sym))
        return 1

    def timg_handler(self, addr, fmt, size, width, height, pal):
        sym = self.create_symbol(
            addr=addr, in_segment=self.in_segment, type="data", reference=True
        )
        gfxd_printf(self.format_sym_name(sym))
        return 1

    def cimg_handler(self, addr, fmt, size, width):
        sym = self.create_symbol(
            addr=addr, in_segment=self.in_segment, type="data", reference=True
        )
        gfxd_printf(self.format_sym_name(sym))
        return 1

    def zimg_handler(self, addr):
        sym = self.create_symbol(
            addr=addr, in_segment=self.in_segment, type="data", reference=True
        )
        gfxd_printf(self.format_sym_name(sym))
        return 1

    def dl_handler(self, addr):
        # Look for 'Gfx'-typed symbols first
        sym = self.retrieve_sym_type(symbols.all_symbols_dict, addr, "Gfx")

        if not sym:
            sym = self.create_symbol(
                addr=addr, in_segment=self.in_segment, type="data", reference=True
            )
        gfxd_printf(self.format_sym_name(sym))
        return 1

    def mtx_handler(self, addr):
        sym = self.create_symbol(
            addr=addr, in_segment=self.in_segment, type="data", reference=True
        )
        gfxd_printf(f"&{self.format_sym_name(sym)}")
        return 1

    def lookat_handler(self, addr, count):
        sym = self.create_symbol(
            addr=addr, in_segment=self.in_segment, type="data", reference=True
        )
        gfxd_printf(self.format_sym_name(sym))
        return 1

    def light_handler(self, addr, count):
        sym = self.create_symbol(
            addr=addr, in_segment=self.in_segment, type="data", reference=True
        )
        gfxd_printf(self.format_sym_name(sym))
        return 1

    def vtx_handler(self, addr, count):
        # Look for 'Vtx'-typed symbols first
        sym = self.retrieve_sym_type(symbols.all_symbols_dict, addr, "Vtx")

        if not sym:
            sym = self.create_symbol(
                addr=addr,
                in_segment=self.in_segment,
                type="Vtx",
                reference=True,
                search_ranges=True,
            )

        index = int((addr - sym.vram_start) / 0x10)
        gfxd_printf(f"&{self.format_sym_name(sym)}[{index}]")
        return 1

    def vp_handler(self, addr):
        sym = self.create_symbol(
            addr=addr, in_segment=self.in_segment, type="data", reference=True
        )
        gfxd_printf(f"&{self.format_sym_name(sym)}")
        return 1

    def macro_fn(self):
        gfxd_puts("    ")
        gfxd_macro_dflt()
        gfxd_puts(",\n")
        return 0

    def disassemble_data(self, rom_bytes):
        assert isinstance(self.rom_start, int)
        assert isinstance(self.rom_end, int)
        assert isinstance(self.vram_start, int)

        gfx_data = rom_bytes[self.rom_start : self.rom_end]
        segment_length = len(gfx_data)
        if (segment_length) % 8 != 0:
            error(
                f"Error: gfx segment {self.name} length ({segment_length}) is not a multiple of 8!"
            )

        out_str = "" if self.data_only else options.opts.generated_c_preamble + "\n\n"

        sym = self.create_symbol(
            addr=self.vram_start, in_segment=True, type="data", define=True
        )

        gfxd_input_buffer(gfx_data)

        # TODO terrible guess at the size we'll need - improve this
        outb = bytes([0] * segment_length * 100)
        outbuf = gfxd_output_buffer(outb, len(outb))

        gfxd_target(self.get_gfxd_target())
        gfxd_endian(
            GfxdEndian.big if options.opts.endianness == "big" else GfxdEndian.little, 4
        )

        # Callbacks
        gfxd_macro_fn(self.macro_fn)

        gfxd_tlut_callback(self.tlut_handler)
        gfxd_timg_callback(self.timg_handler)
        gfxd_cimg_callback(self.cimg_handler)
        gfxd_zimg_callback(self.zimg_handler)
        gfxd_dl_callback(self.dl_handler)
        gfxd_mtx_callback(self.mtx_handler)
        gfxd_lookat_callback(self.lookat_handler)
        gfxd_light_callback(self.light_handler)
        # gfxd_seg_callback ?
        gfxd_vtx_callback(self.vtx_handler)
        gfxd_vp_callback(self.vp_handler)
        # gfxd_uctext_callback ?
        # gfxd_ucdata_callback ?
        # gfxd_dram_callback ?

        gfxd_execute()

        if self.data_only:
            out_str += gfxd_buffer_to_string(outbuf)
        else:
            out_str += "Gfx " + self.format_sym_name(sym) + "[] = {\n"
            out_str += gfxd_buffer_to_string(outbuf)
            out_str += "};\n"

        # Poor man's light fix until we get my libgfxd PR merged
        def light_sub_func(match):
            light = match.group(0)
            addr = int(light[12:], 0)
            sym = self.create_symbol(
                addr=addr, in_segment=self.in_segment, type="data", reference=True
            )
            return self.format_sym_name(sym)

        out_str = re.sub(LIGHTS_RE, light_sub_func, out_str)

        return out_str

    def split(self, rom_bytes: bytes):
        out_path = self.out_path()
        if self.file_text and out_path:
            out_path.parent.mkdir(parents=True, exist_ok=True)

            with out_path.open("w", encoding="utf-8", newline="\n") as f:
                f.write(self.file_text)

    def should_scan(self) -> bool:
        return (
            options.opts.is_mode_active("gfx")
            and self.rom_start is not None
            and self.rom_end is not None
        )

    def should_split(self) -> bool:
        return self.extract and options.opts.is_mode_active("gfx")

    @staticmethod
    def estimate_size(yaml: Union[Dict, List]) -> Optional[int]:
        if isinstance(yaml, dict) and "length" in yaml:
            return yaml["length"] * 0x10
        return None



import crunch64

from .decompressor import CommonSegDecompressor


class N64SegYay0(CommonSegDecompressor):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def compression_type(self):
        return "Yay0"

    def decompress(self, compressed_bytes: bytes) -> bytes:
        return crunch64.yay0.decompress(compressed_bytes)



from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional

from . import log


@dataclass(frozen=True)
class VramClass:
    name: str
    vram: int
    given_vram_symbol: Optional[str] = None
    follows_classes: List[str] = field(default_factory=list, compare=False)

    @property
    def vram_symbol(self) -> Optional[str]:
        if self.given_vram_symbol is not None:
            return self.given_vram_symbol
        elif self.follows_classes:
            return self.name + "_CLASS_VRAM"
        else:
            return None


_vram_classes: Dict[str, VramClass] = {}


def initialize(yaml: Any):
    global _vram_classes

    _vram_classes = {}

    if yaml is None:
        return

    if not isinstance(yaml, list):
        log.error("vram_classes must be a list")

    class_names = set()
    for vram_class in yaml:
        if isinstance(vram_class, dict):
            if "name" not in vram_class:
                log.error(f"vram_class ({vram_class}) must have a name")
            class_names.add(vram_class["name"])
        elif isinstance(vram_class, list):
            class_names.add(vram_class[0])

    for vram_class in yaml:
        name: str
        vram: int
        vram_symbol: Optional[str] = None
        follows_classes: List[str] = []

        if isinstance(vram_class, dict):
            if "name" not in vram_class:
                log.error(f"vram_class ({vram_class}) must have a name")
            name = vram_class["name"]

            if "vram" not in vram_class:
                log.error(f"vram_class ({vram_class}) must have a vram")
            vram = vram_class["vram"]

            if "vram_symbol" in vram_class:
                vram_symbol = vram_class["vram_symbol"]
                if not isinstance(vram_symbol, str):
                    log.error(
                        f"vram_symbol ({vram_symbol})must be a string, got {type(vram_symbol)}"
                    )

            if "follows_classes" in vram_class:
                follows_classes = vram_class["follows_classes"]
                if not isinstance(follows_classes, list):
                    log.error(
                        f"vram_symbol ({follows_classes})must be a list, got {type(follows_classes)}"
                    )
                for follows_class in follows_classes:
                    if follows_class not in class_names:
                        log.error(
                            f"follows_class ({follows_class}) not found in vram_classes"
                        )
        elif isinstance(vram_class, list):
            if len(vram_class) != 2:
                log.error(
                    f"vram_class ({vram_class}) must have 2 elements, got {len(vram_class)}"
                )
            name = vram_class[0]
            vram = vram_class[1]
        else:
            log.error(f"vram_class must be a dict or list, got {type(vram_class)}")

        if not isinstance(name, str):
            log.error(f"vram_class name ({name}) must be a string, got {type(name)}")
        if not isinstance(vram, int):
            log.error(f"vram_class vram ({vram}) must be an int, got {type(vram)}")
        if name in _vram_classes:
            log.error(f"Duplicate vram class name '{name}'")
        _vram_classes[name] = VramClass(name, vram, vram_symbol, follows_classes)


def resolve(name: str) -> VramClass:
    if name not in _vram_classes:
        log.error(f"Unknown vram class '{name}'")
    return _vram_classes[name]



from . import asm as asm
from . import bin as bin
from . import bss as bss
from . import code as code
from . import codesubsegment as codesubsegment
from . import cpp as cpp
from . import c as c
from . import databin as databin
from . import data as data
from . import group as group
from . import hasm as hasm
from . import header as header
from . import lib as lib
from . import linker_offset as linker_offset
from . import rdata as rdata
from . import rodatabin as rodatabin
from . import rodata as rodata
from . import sbss as sbss
from . import sdata as sdata
from . import segment as segment
from . import textbin as textbin



from pathlib import Path
from typing import List, TYPE_CHECKING

from ...util import log, options

from .img import N64SegImg

if TYPE_CHECKING:
    from .palette import N64SegPalette


# Base class for CI4/CI8
class N64SegCi(N64SegImg):
    def parse_palette_names(self, yaml, args) -> List[str]:
        ret = [self.name]
        if isinstance(yaml, dict):
            if "palettes" in yaml:
                ret = yaml["palettes"]
        elif len(args) > 2:
            ret = args[2]

        if isinstance(ret, str):
            ret = [ret]
        return ret

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.palettes: "List[N64SegPalette]" = []
        self.palette_names = self.parse_palette_names(self.yaml, self.args)

    def scan(self, rom_bytes: bytes) -> None:
        self.n64img.data = rom_bytes[self.rom_start : self.rom_end]

    def out_path_pal(self, pal_name) -> Path:
        type_extension = f".{self.type}" if options.opts.image_type_in_extension else ""

        if len(self.palettes) == 1:
            # If there's only one palette, use the ci name
            out_name = self.name
        elif pal_name.startswith(self.name):
            # Otherwise, if the palette name starts with / equals the ci name, use that
            out_name = pal_name
        else:
            # Otherwise, just append the palette name to the ci name
            out_name = f"{self.name}_{pal_name}"

        return options.opts.asset_path / self.dir / f"{out_name}{type_extension}.png"

    def split(self, rom_bytes):
        self.check_len()

        assert self.palettes is not None
        if len(self.palettes) == 0:
            # TODO: output with blank palette
            log.error(
                f"no palettes have been mapped to ci segment `{self.name}`\n(hint: add a palette segment with the same name or use the `palettes:` field of this segment to specify palettes by name')"
            )

        assert isinstance(self.rom_start, int)
        assert isinstance(self.rom_end, int)
        self.n64img.data = rom_bytes[self.rom_start : self.rom_end]

        for palette in self.palettes:
            path = self.out_path_pal(palette.name)
            path.parent.mkdir(parents=True, exist_ok=True)

            self.n64img.palette = palette.parse_palette(rom_bytes)
            self.n64img.write(path)

            self.log(f"Wrote {path.name} to {path}")



#! /usr/bin/env python3

from __future__ import annotations

import dataclasses
from pathlib import Path
import spimdisasm
from spimdisasm.elf32 import (
    Elf32File,
    Elf32Constants,
    Elf32SectionHeaderFlag,
    Elf32ObjectFileType,
)
from typing import Optional

from .. import log


ELF_SECTION_MAPPING: dict[str, str] = {
    ".text": "asm",
    ".data": "data",
    ".rodata": "rodata",
    ".bss": "bss",
    ".sdata": "sdata",
    ".sbss": "sbss",
    ".gcc_except_table": "gcc_except_table",
    ".lit4": "lit4",
    ".lit8": "lit8",
    ".ctor": "ctor",
    ".vtables": "vtables",
    ".vutext": "textbin",  # No "proper" support yet
    ".vudata": "databin",  # No "proper" support yet
}

# Section to not put into the elf_section_names list, because splat doesn't
# know have support for them yet.
ELF_SECTIONS_IGNORE: set[str] = {
    ".vutext",
    ".vudata",
    ".vubss",
}

ELF_SMALL_SECTIONS: set[str] = {
    ".lit4",
    ".lit8",
    ".sdata",
    ".srdata",
    ".sbss",
}


@dataclasses.dataclass
class Ps2Elf:
    entrypoint: int
    segs: list[FakeSegment]
    size: int
    compiler: str
    elf_section_names: list[tuple[str, bool]]
    gp: Optional[int]
    ld_gp_expression: Optional[str]

    @staticmethod
    def get_info(elf_path: Path, elf_bytes: bytes) -> Optional[Ps2Elf]:
        # Avoid spimdisasm from complaining about unknown sections.
        spimdisasm.common.GlobalConfig.QUIET = True

        elf = Elf32File(elf_bytes)
        if elf.header.type != Elf32ObjectFileType.EXEC.value:
            log.write("Elf file is not an EXEC type.", status="warn")
            return None
        if elf.header.machine != 8:
            # 8 corresponds to EM_MIPS
            # We only care about mips binaries.
            log.write("Elf file is not a MIPS binary.", status="warn")
            return None
        if Elf32Constants.Elf32HeaderFlag._5900 not in elf.elfFlags:
            log.write("Missing 5900 flag", status="warn")
            return None

        if elf.reginfo is not None and elf.reginfo.gpValue != 0:
            gp = elf.reginfo.gpValue
        else:
            gp = None
        first_small_section_info: Optional[tuple[str, int]] = None

        first_segment_name = "cod"
        segs = [FakeSegment(first_segment_name, 0, 0, [])]

        # TODO: check `.comment` section for any compiler info
        compiler = "EEGCC"

        elf_section_names: list[tuple[str, bool]] = []

        first_offset: Optional[int] = None
        rom_size = 0

        previous_type = Elf32Constants.Elf32SectionHeaderType.PROGBITS
        do_new_segs = False

        # Loop over all the sections.
        # Treat every normal elf section as part as the "main" segment.
        # If we see a PROGBITS after a NOBITS then we shift the behavior into
        # putting every new elf section into their own segment, this can happen
        # when the elf contains a few special sections like `.mfifo`.
        section_headers = sorted(elf.sectionHeaders, key=lambda x: x.offset)
        for section in section_headers:
            if section.size == 0:
                # Skip over empty sections
                continue

            name = elf.shstrtab[section.name]
            if name == ".mwcats":
                compiler = "MWCCPS2"
                continue

            flags, _unknown_flags = Elf32SectionHeaderFlag.parseFlags(section.flags)
            # if _unknown_flags != 0:
            #     print(name, f"0x{_unknown_flags:08X}")
            if Elf32SectionHeaderFlag.ALLOC not in flags:
                # We don't care about non-alloc sections
                continue

            # We want PROGBITS (actual data) and NOBITS (bss and similar)
            typ = Elf32Constants.Elf32SectionHeaderType.fromValue(section.type)
            is_nobits = typ == Elf32Constants.Elf32SectionHeaderType.NOBITS
            if typ == Elf32Constants.Elf32SectionHeaderType.PROGBITS:
                if previous_type == Elf32Constants.Elf32SectionHeaderType.NOBITS:
                    do_new_segs = True
                pass
            elif typ == Elf32Constants.Elf32SectionHeaderType.NOBITS:
                pass
            elif typ == Elf32Constants.Elf32SectionHeaderType.MIPS_REGINFO:
                continue
            else:
                log.write(
                    f"Unknown section type '{typ}' ({name}) found in the elf",
                    status="warn",
                )
                return None

            if first_offset is None:
                first_offset = section.offset

            if first_small_section_info is None and name in ELF_SMALL_SECTIONS:
                first_small_section_info = (name, section.addr)

            start = section.offset - first_offset
            size = section.size

            if do_new_segs:
                segs.append(FakeSegment(name.lstrip("."), 0, start, []))

            # Try to map this section to something splat can understand.
            splat_segment_type = ELF_SECTION_MAPPING.get(name)
            if splat_segment_type is None:
                # Let's infer based on the section's flags
                if is_nobits:
                    splat_segment_type = "bss"
                elif Elf32SectionHeaderFlag.EXECINSTR in flags:
                    splat_segment_type = "asm"
                elif Elf32SectionHeaderFlag.WRITE in flags:
                    splat_segment_type = "data"
                else:
                    # Whatever...
                    splat_segment_type = "rodata"

            if name.startswith("."):
                valid_for_splat = (
                    name in ELF_SECTION_MAPPING
                    and name not in ELF_SECTIONS_IGNORE
                    and not do_new_segs
                )
                elf_section_names.append((name, valid_for_splat))

            new_section = ElfSection(
                name,
                splat_segment_type,
                section.addr,
                start,
                size,
                is_nobits,
            )
            segs[-1].sections.append(new_section)
            if is_nobits:
                segs[-1].bss_size += size
            else:
                rom_size = start + size

            previous_type = typ

        # There are some games where they just squashed most sections into a
        # single one, making the elf_section_names list pretty useless.
        # We try to detect this and provide a default list if that's the case,
        # hoping for the best.
        if len(elf_section_names) < 4:
            elf_section_names = [
                (".text", True),
                (".vutext", False),
                (".data", True),
                (".vudata", False),
                (".rodata", True),
                (".gcc_except_table", True),
                (".lit8", True),
                (".lit4", True),
                (".sdata", True),
                (".sbss", True),
                (".bss", True),
                (".vubss", False),
            ]

        # Fixup vram address of segments
        for seg in segs:
            seg.vram = seg.sections[0].vram

        ld_gp_expression = None
        if gp is not None and first_small_section_info is not None:
            section_name, section_address = first_small_section_info
            if gp > section_address:
                diff = gp - section_address
                ld_gp_expression = f"{first_segment_name}_{section_name.strip('.').upper()}_START + 0x{diff:04X}"

        return Ps2Elf(
            elf.header.entry,
            segs,
            rom_size,
            compiler,
            elf_section_names,
            gp,
            ld_gp_expression,
        )


@dataclasses.dataclass
class FakeSegment:
    name: str
    vram: int
    start: int
    sections: list[ElfSection]
    bss_size: int = 0


@dataclasses.dataclass
class ElfSection:
    name: str
    splat_segment_type: str
    vram: int
    start: int
    size: int
    is_nobits: bool
